#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to update the names of classifiers in every original csv input
file. Replace '_1' and '_2' with the real classifier names.
"""

# Built-in modules #

# Third party modules #
from tqdm import tqdm

# First party modules #
from autopaths import Path

# Internal modules #
from eu_cbm_hat.core.continent import continent

###############################################################################
class OrigClassifRenamer(object):
    """
    `_1` becomes `forest_type` etc.
    """

    mapping = {
        '_1' : 'status',
        '_2' : 'forest_type',
        '_3' : 'region',
        '_4' : 'mgmt_type',
        '_5' : 'mgmt_strategy',
        '_6' : 'climate',
        '_7' : 'con_broad',
        '_8' : 'site_index',
        '_9' : 'growth_period',
    }

    def __init__(self, country):
        # Default attributes #
        self.country = country

    csv_list = ['events.csv', 'inventory.csv', 'transitions.csv',
                'growth_curves.csv']

    def __call__(self):
        for item in self.csv_list:
            csv_path = Path(self.country.data_dir + 'orig/csv/' + item)
            self.rename(csv_path)

    #------------------------------- Methods ---------------------------------#
    def rename(self, csv_path):
        """Loads one CSV files and renames columns."""
        # Load from disk #
        header = csv_path.first
        header = header.split(',')
        # Modify #
        header = map(self.mapping.get, header, header)
        # Write to disk #
        header = ','.join(header)
        csv_path.remove_first_line()
        csv_path.prepend(header)

###############################################################################
if __name__ == '__main__':
    renamers = [OrigClassifRenamer(c) for c in continent]
    for renamer in tqdm(renamers): renamer()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #

# Internal modules #
from eu_cbm_hat.pump.column_order import events_cols

###############################################################################
#------------------------ Dataframe conversions --------------------------#
def events_wide_to_long(country, events):
    """Reshape disturbance events from wide to long format."""
    # Make a copy of the dataframe #
    df = events.copy()
    # We want to pivot on all columns except two #
    skip  = ['step', 'amount']
    cols  = [col for col in events_cols if col not in skip]
    # Check wide column names are within the range
    year_range = range(1900, 2101)
    amount_list = [f"amount_{year}" for year in year_range]
    cols_not_in_range = set(events.columns) - set(cols) - set(amount_list)
    if cols_not_in_range:
        msg = f"Column name {cols_not_in_range} in the events input data "
        msg += f"is not in the {year_range}."
        raise ValueError(msg)
    # Reshape from wide to long format #
    try:
        df = pandas.wide_to_long(df,
                                 stubnames = "amount",
                                 i         = cols,
                                 j         = "year",
                                 sep       = '_')
    # Better error message #
    except ValueError as error:
        if 'uniquely identify each row' not in str(error): raise
        msg = "You probably have two or more rows in your events file " \
              "which both have the same classifier values. Hence one " \
              "cannot convert it from wide to long format."
        raise ValueError(msg) from error
    # Drop rows that don't have an amount #
    df = df.dropna()
    # Reset index #
    df = df.reset_index()
    # Convert years to time steps #
    df['step'] = country.year_to_timestep(df['year'])
    # Drop the year column #
    df = df.drop(columns=['year'])
    # Reorder columns according to the correct input order #
    df = df[events_cols]
    # Return #
    return df

def events_long_to_wide(country, events):
    """Reshape disturbance events from long to wide format."""
    # Make a copy of the dataframe #
    df = events.copy()
    # Convert the step to years #
    df['year'] = country.timestep_to_year(df['step'])
    # Drop the steps column #
    df = df.drop(columns=['step'])
    # We want to pivot on all columns except two #
    skip_cols = ['step', 'amount']
    cols = [col for col in events_cols if col not in skip_cols]
    # Reshape from long to wide format #
    df = df.pivot(index   = cols,
                  columns = 'year',
                  values  = 'amount')
    # We don't want the columns to be called 'year' though #
    df = df.rename_axis(columns=None)
    # Add 'amount_' in front of every column name #
    df = df.add_prefix('amount_')
    # Remove rows that are all NaNs #
    df = df.dropna(how='all')
    # Reset index #
    df = df.reset_index()
    # Sort entries #
    df = df.sort_values(cols)
    # Add the scenario column #
    df.insert(0, 'scenario', 'reference')
    # Return #
    return df



""" Run the EU_CBM_HAT scenario combinations

Usage:

    cd ~/repos/eu_cbm/eu_cbm_hat/scripts/running/
    # or on BDAP
    cd /storage/$USER/eu_cbm_hat/scripts/running
    ipython -i run_scenario_combo.py -- --combo_name pikbau --last_year 2070
    ipython -i run_scenario_combo.py -- --combo_name pikfair --last_year 2070
    # Only run a few countries
    ipython -i run_scenario_combo.py -- --combo_name pikbau --last_year 2070 --countries LU CZ

The version of p_umap in eu_cbm_hat/combos/base_combo.py Combination.__call__()
method was using a function that takes a list of runner objects as one its
argument. The problem was that `p_umap(run_country, runner_items, num_cpus=4)`
lead to `TypeError: cannot pickle 'sqlite3.Connection' object`. This is because
SQLite connections cannot be pickled and sent to subprocesses. Instead of
passing a runner as a function argument, the solution is to pass only the
country code as a function argument and have the function create the runner
object inside the sub-process directly.

"""

import argparse
from p_tqdm import p_umap
from eu_cbm_hat.core.continent import continent

def run_country(args):
    """Run a single country, only based on the code as input"""
    combo_name, last_year, country_code = args
    runner = continent.combos[combo_name].runners[country_code][-1]
    runner.num_timesteps = last_year - runner.country.inventory_start_year
    try:
        runner.run()
    except Exception as e:
        print(e)

parser = argparse.ArgumentParser(description='Run the EU_CBM_HAT scenario combinations')
parser.add_argument('--combo_name', type=str, help='Name of the scenario combo to be run')
parser.add_argument('--last_year', type=int, help='Last year to be simulated')
parser.add_argument('--countries', nargs='+', default=None, help='List of country ISO2 codes')

args = parser.parse_args()

LAST_YEAR = args.last_year
COMBO_NAME = args.combo_name
COUNTRIES = args.countries

# If no countries are specified, run for all countries
if COUNTRIES is None:
    COUNTRIES = list(continent.combos[COMBO_NAME].runners.keys())

runner_items = [(COMBO_NAME, LAST_YEAR, k) for k in COUNTRIES]
result = p_umap(run_country, runner_items, num_cpus=4)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from autopaths.auto_paths import AutoPaths
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.pump.long_or_wide import events_wide_to_long

###############################################################################
class OrigData(object):
    """
    This class will provide access to the original data of a Country
    as several pandas data frames.

    The data content itself for the historical period was taken from the
    original `cbmcfs3` dataset composed by Roberto P.

    To copy all the data from the `cbmcfs3_data` repository do the following:

        >>> from eu_cbm_hat.core.continent import continent
        >>> for country in continent: country.orig_data.copy_from_cbmcfs3()

    This was done at some point before the data was reorganized and changed
    by Viorel B. It can still be seen in `f42fb77` and it depends on the
    `cbmcfs3_runner` python module to be run.
    """

    all_paths = """
    /config/associations.csv               # Static
    /common/age_classes.csv                # Static
    /common/classifiers.csv                # Static
    /common/disturbance_types.csv          # Static
    /silv/irw_frac_by_dist.csv             # Has scenario column
    /silv/vol_to_mass_coefs.csv            # Has scenario column
    /silv/events_templates.csv             # Has scenario column
    /silv/harvest_factors.csv              # Has scenario column
    /activities/                           # Dynamic
    """

    # These files are not listed in the paths above because they depend on
    # different activities.
    files_to_be_generated = ['growth_curves',
                             'transitions',
                             'events',
                             'inventory']

    def __init__(self, parent):
        # Default attributes #
        self.parent  = parent
        self.country = parent
        # Directories #
        self.paths = AutoPaths(self.parent.data_dir, self.all_paths)

    def __getitem__(self, item):
        # A single argument is used for the static files #
        if isinstance(item, str):
            return pandas.read_csv(str(self.paths[item]), dtype=str)
        # Two arguments as a tuple indicates an activity #
        elif len(item) == 2:
            activity, file = item
            path = self.paths.activities_dir + activity + '/' + file + '.csv'
            return pandas.read_csv(str(path), dtype=str)
        # Wrong number of arguments #
        raise Exception("Undefined input data access '%s'." % str(item))

    #----------------------------- Properties --------------------------------#
    @property_cached
    def activities(self):
        """All the activities that exist for the current country."""
        return [d.name for d in self.paths.activities_dir.flat_directories]

    @property_cached
    def classif_names(self):
        # Load #
        result = self['classifiers']
        # Query #
        result = result.query("classifier_value_id == '_CLASSIFIER'")
        # Get a series #
        result = result.set_index('classifier_number')['name']
        # Link number to name #
        result = {'_' + str(k): v for k,v in result.items()}
        # Return #
        return result

    #------------------------------- Methods ---------------------------------#
    def load(self, name, clfrs_names=False, to_long=False):
        """
        Loads one of the dataframes in the orig data and adds information
        to it.
        """
        # Load from CSV #
        df = self[name]
        # Optionally rename classifiers #
        if clfrs_names: df = df.rename(columns=self.classif_names)
        # Optionally convert to the long format #
        if to_long: df = events_wide_to_long(self.country, df)
        # Return #
        return df


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import os

# First party modules #
from autopaths.dir_path   import DirectoryPath
from autopaths.auto_paths import AutoPaths
from plumbing.cache       import property_cached

# Internal modules #
from eu_cbm_hat import eu_cbm_aidb_dir


###############################################################################
class AIDB(object):
    """
    This class will provide access to the archive index database
    also called 'cbm_defaults' in libcbm.
    It is an SQLite3 database that weighs approx 18 MiB.

    To symlink the single test database to all countries do the following:

        >>> from eu_cbm_hat.core.continent import continent
        >>> for country in continent: country.aidb.symlink_single_aidb()

    To symlink every AIDB from every countries do the following:

        >>> from eu_cbm_hat.core.continent import continent
        >>> for country in continent: country.aidb.symlink_all_aidb()
    """

    all_paths = """
    /config/aidb.db
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Directories #
        self.paths = AutoPaths(self.parent.data_dir, self.all_paths)
        # Path to the database in a separate repository #
        self.repo_file = eu_cbm_aidb_dir + 'countries/' + self.parent.iso2_code \
                         + '/aidb.db'

    def __bool__(self):
        return bool(self.paths.aidb)

    #----------------------------- Properties --------------------------------#
    @property_cached
    def db(self):
        """
        Returns a `plumbing.databases.sqlite_database.SQLiteDatabase` object
        useful for reading and modifying entries and tables.

        In addition one can also read/write to the AIDB files easily:

            >>> df = country.aidb.db.read_df('species')

        To overwrite a table with a df:

            >>> country.aidb.db.write_df(df, 'species')
        """
        from plumbing.databases.sqlite_database import SQLiteDatabase
        return SQLiteDatabase(self.paths.aidb)

    @property
    def vol_conv_to_biomass(self):
        """Volume to biomass conversion factors

        Example:

            >>>> from eu_cbm_hat.core.continent import continent
            >>>> r = continent.combos['hat'].runners['ZZ'][-1]
            >>>> r.country.aidb.vol_conv_to_biomass

        """
        #load table with factors
        vol_to_biomass_factors = (self.db.read_df('vol_to_bio_factor')
                                  .drop_duplicates(subset=['a', 'b'], keep='last'))
        #load the parameteres a and b for species from aidb
        vol_to_bio_f = vol_to_biomass_factors.rename(columns = {'id':'vol_to_bio_factor_id'})
        vol_to_bio_sp = (self.db.read_df('vol_to_bio_species')
                         .drop_duplicates(subset=['species_id'], keep='last')
                        )
        # add species ids to a and b values
        vol_to_bio_sp_f = vol_to_bio_sp.merge(vol_to_bio_f, how = 'inner')
        species_name = self.db.read_df('species_tr')
        sp_name = species_name.rename(columns = {'name': 'cbm_forest_name'})
        # get the final database on spatial_units_ids and species or forest types
        vol_to_bio_species_factors = (vol_to_bio_sp_f
                                      .merge(sp_name, how = 'inner', on = 'species_id')
                                      .drop(columns = 'vol_to_bio_factor_id')
                                      .rename(columns = {'species_name':'forest_type'})                                                                )
        #select only the relevant columns
        colns = ['cbm_forest_name','spatial_unit_id', 'species_id', 'a', 'b']
        cbm_biom = vol_to_bio_species_factors[colns]
        return cbm_biom


    #------------------------------- Methods ---------------------------------#
    def symlink_single_aidb(self):
        """
        During development, and for testing purposes we have a single AIDB
        that all countries can share and that is found in another repository.
        """
        # The path to the SQLite3 file #
        source = DirectoryPath(eu_cbm_aidb_dir + 'aidb.db')
        # Check it exists #
        try:
            assert source
        except AssertionError:
            msg = "The sqlite3 database at '%s' does not seems to exist."
            raise AssertionError(msg % source)
        # Symlink #
        destin = self.paths.aidb
        source.link_to(destin)

    def symlink_all_aidb(self):
        """In production, every country has its own AIDB."""
        # Check the AIDB exists #
        try:
            assert self.repo_file
        except AssertionError:
            msg = "The sqlite3 database at '%s' does not seems to exist."
            raise AssertionError(msg % self.repo_file)
        # The destination #
        destin = self.paths.aidb
        # Remove destination if it already exists #
        destin.remove()
        # Symlink #
        self.repo_file.link_to(destin)
        # Return #
        return 'Symlink success for ' + self.parent.iso2_code + '.'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #

# First party modules #

# Internal modules #

###############################################################################
# We need to record the column order in which data should appear in the
# `events.csv` file.
events_cols = [
    'status',
    'forest_type',
    'region',
    'mgmt_type',
    'mgmt_strategy',
    'climate',
    'con_broad',
    'site_index',
    'growth_period',
    'using_id',
    'sw_start',
    'sw_end',
    'hw_start',
    'hw_end',
    'min_since_last_dist',
    'max_since_last_dist',
    'last_dist_id',
    'min_tot_biom_c',
    'max_tot_biom_c',
    'min_merch_soft_biom_c',
    'max_merch_soft_biom_c',
    'min_merch_hard_biom_c',
    'max_merch_hard_biom_c',
    'min_tot_stem_snag_c',
    'max_tot_stem_snag_c',
    'min_tot_soft_stem_snag_c',
    'max_tot_soft_stem_snag_c',
    'min_tot_hard_stem_snag_c',
    'max_tot_hard_stem_snag_c',
    'min_tot_merch_stem_snag_c',
    'max_tot_merch_stem_snag_c',
    'min_tot_merch_soft_stem_snag_c',
    'max_tot_merch_soft_stem_snag_c',
    'min_tot_merch_hard_stem_snag_c',
    'max_tot_merch_hard_stem_snag_c',
    'efficiency',
    'sort_type',
    'measurement_type',
    'amount',
    'dist_type_name',
    'step',
]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to harmonize the names of classifiers in every country to snake_case.
"""

# Built-in modules #

# Third party modules #
import pandas
from tqdm import tqdm

# First party modules #

# Internal modules #
from eu_cbm_hat.core.continent import continent

###############################################################################
class ClassifierRenamer(object):
    """
    `Forest type` becomes `forest_type` etc.
    """

    mapping = {
        'Status'                         : 'status',
        'Forest type'                    : 'forest_type',
        'Region'                         : 'region',
        'Management type'                : 'mgmt_type',
        'Management strategy'            : 'mgmt_strategy',
        'Climatic unit'                  : 'climate',
        'Conifers/Broadleaves'           : 'con_broad',
        'Site index'                     : 'site_index',
        'Simulation period (for yields)' : 'growth_period',
    }

    def __init__(self, country):
        # Default attributes #
        self.country = country

    def __call__(self):
        # Get path #
        path = self.country.orig_data.paths.classifiers
        # Load dataframe #
        df = pandas.read_csv(str(path))
        # Modify dataframe #
        df['name'] = df['name'].replace(self.mapping)
        # Write dataframe back to disk #
        df.to_csv(str(path), index=False, float_format='%g')

###############################################################################
if __name__ == '__main__':
    renamers = [ClassifierRenamer(c) for c in continent]
    for renamer in tqdm(renamers): renamer()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
from libcbm.input.sit import sit_cbm_factory
from libcbm.model.cbm import cbm_simulator

# First party modules #

# Internal modules #

# Constants #
create_proc = sit_cbm_factory.create_sit_rule_based_processor
create_func = cbm_simulator.create_in_memory_reporting_func

###############################################################################
class Simulation(object):
    """This class will run a `libcbm_py` simulation."""

    def __init__(self, parent):
        # Default attributes #
        self.parent  = parent
        self.runner  = parent
        self.country = self.runner.country
        # Record if we ended with an error or not #
        self.error = None

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    #--------------------------- Special Methods -----------------------------#
    def switch_period(self, cbm_vars):
        """
        If t=1, we know this is the first timestep, and nothing has yet been
        done to the post-spinup pools. It is at this moment that we want to
        change the growth curves, and this can be done by switching the
        classifier value of each inventory record.
        """
        # Print message #
        msg = "Carbon pool initialization period is finished." \
              " Now starting the `current` period.\n"
        self.parent.log.info(msg)
        # The name of our extra classifier #
        key = 'growth_period'
        # The value that the classifier should take for all timesteps #
        val = "Cur"
        # Get the corresponding ID in the libcbm simulation #
        id_of_cur = self.sit.classifier_value_ids[key][val]
        # Modify the whole column of the dataframe #
        cbm_vars.classifiers[key] = id_of_cur
        # Return #
        return cbm_vars

    def dynamics_func(self, timestep, cbm_vars):
        """
        See the simulate method of the `libcbm_py` simulator:

            https://github.com/cat-cfs/libcbm_py/blob/master/libcbm/
            model/cbm/cbm_simulator.py#L148
        """
        # Check if we want to switch growth period #
        if timestep == 1: cbm_vars = self.switch_period(cbm_vars)
        # Print a message #
        self.parent.log.info(f"Time step {timestep} is about to run.")
        # Run the usual rule based processor #
        cbm_vars = self.rule_based_proc.pre_dynamics_func(timestep, cbm_vars)
        # Return #
        return cbm_vars

    #------------------------------- Methods ---------------------------------#
    # noinspection PyBroadException
    def __call__(self, interrupt_on_error=False):
        """
        Wrap the `run()` method by catching any type of exception
        and logging it. This is useful when running all countries one after
        each other and not wanting that process to be interrupted even if a few
        countries fail along the way.
        """
        try:
            self.run()
        except Exception:
            message = "Runner '%s' encountered an exception. See log file."
            self.runner.log.error(message % self.runner.short_name)
            self.runner.log.exception("Exception", exc_info=True)
            self.error = True
            if interrupt_on_error: raise

    def run(self):
        """
        Call `libcbm_py` to run the actual CBM simulation after creating some
        objects.
        The interaction with `libcbm_py` is decomposed in several calls to pass
        a `.json` config, a default database (also called aidb) and csv files.
        """
        # Message #
        self.runner.log.info("Setting up the libcbm_py objects.")
        # The 'AIDB' path as it was called previously #
        db_path = self.runner.country.aidb.paths.db
        if not db_path:
            msg = "The database file at '%s' was not found."
            raise FileNotFoundError(msg % self.runner.country.aidb.paths.db)
        # Create a SIT object #
        json_path = str(self.runner.paths.json)
        self.sit = sit_cbm_factory.load_sit(json_path, str(db_path))
        # Do some initialization #
        init_inv = sit_cbm_factory.initialize_inventory
        self.clfrs, self.inv = init_inv(self.sit)
        # This will contain results #
        self.results, self.reporting_func = create_func()
        # Create a CBM object #
        with sit_cbm_factory.initialize_cbm(self.sit) as self.cbm:
            # Create a function to apply rule based events #
            self.rule_based_proc = create_proc(self.sit, self.cbm)
            # Message #
            self.runner.log.info("Calling the cbm_simulator.")
            # Run #
            cbm_simulator.simulate(
                self.cbm,
                n_steps           = self.runner.num_timesteps,
                classifiers       = self.clfrs,
                inventory         = self.inv,
                pre_dynamics_func = self.dynamics_func,
                reporting_func    = self.reporting_func
            )
        # If we got here then we did not encounter any simulation error #
        self.error = False
        # Return for convenience #
        return self.results

    def clear(self):
        """
        Remove all objects from RAM otherwise the kernel will kill the python
        process after a couple countries being run.
        """
        if hasattr(self, 'cbm'):            del self.cbm
        if hasattr(self, 'sit'):            del self.sit
        if hasattr(self, 'clfrs'):          del self.clfrs
        if hasattr(self, 'inv'):            del self.inv
        if hasattr(self, 'results'):        del self.results
        if hasattr(self, 'reporting_func'): del self.reporting_func



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from plumbing.cache import property_cached

# Internal modules #

###############################################################################
class Associations(object):
    """
    This class parses the file "associations.csv" and returns
    a dictionary useful to produce the JSON for consumption by
    libcbm.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent

    #----------------------------- Properties --------------------------------#
    @property_cached
    def df(self):
        """
        Load the CSV that is the original 'associations.csv' from cbmcfs3.
        The path is taken from the OrigData class.
        """
        return pandas.read_csv(str(self.parent.orig_data.paths.associations))

    @property_cached
    def all_mappings(self):
        """Return a dictionary for creation of the JSON file."""
        return {
           'map_admin_bound': self.rows_to_list('MapAdminBoundary',
                                                'user_admin_boundary',
                                                'default_admin_boundary'),
           'map_eco_bound':   self.rows_to_list('MapEcoBoundary',
                                                'user_eco_boundary',
                                                'default_eco_boundary'),
           'map_species':     self.rows_to_list('MapSpecies',
                                                'user_species',
                                                'default_species'),
           'map_disturbance': self.rows_to_list('MapDisturbanceType',
                                                'user_dist_type',
                                                'default_dist_type'),
        }

    #------------------------------- Methods ---------------------------------#
    def key_to_rows(self, mapping_name):
        """
        The *mapping_name* has to be one of "keys".
        Here is an example call:

        >>> self.key_to_rows('MapDisturbanceType')

        {'10% commercial thinning': '10% Commercial thinning',
         'Deforestation':           'Deforestation',
         'Fire':                    'Wild Fire',
         'Generic 15%':             'generic 15% mortality',
         'Generic 20%':             'generic 20% mortality',
         'Generic 30%':             'generic 30% mortality'}
        """
        # The query to filter results #
        query = "category == '%s'" % mapping_name
        # Run query #
        df = self.df.query(query).set_index('name_input')
        # Get names that match the AIDB #
        mapping = df['name_aidb'].to_dict()
        # Return #
        return mapping

    def rows_to_list(self, mapping_name, user, default):
        """
        Create a list string by picking the appropriate rows in the CSV file.

        Here is an example call:

        >>> self.rows_to_list('MapSpecies', 'user_species', 'default_species')

        {...}
        """
        return [{user: k, default: v}
                for k, v in self.key_to_rows(mapping_name).items()]




"""
Test salvage silviculture methods

Execute the test suite from bash with py.test as follows:

    cd ~/repos/eu_cbm/eu_cbm_hat/eu_cbm_hat
    pytest

"""

import numpy as np
import pandas
import pytest
from eu_cbm_hat.core.continent import continent
combo  = continent.combos['hat']
runner = combo.runners['ZZ'][-1]

harvest_factors_1 = {'scenario': ['reference', 'reference', 'reference', 'no_skew', 'no_skew'],
                     'product_created': ['irw_and_fw', 'irw_and_fw', 'fw_only', 'irw_and_fw', 'fw_only'],
                     'forest_type': np.nan,
                     'silv_practice': np.nan,
                     'mgmt_type': np.nan,
                     'con_broad': ['con', 'broad', np.nan, np.nan, np.nan],
                     'disturbance_type': np.nan,
                     'value_2015': [0.3, 0.7, 1.0, 1.0, 1.0]}
harvest_factors_1 = pandas.DataFrame(harvest_factors_1)

harvest_factors_2 = harvest_factors_1.copy()
harvest_factors_2["con_broad"] = ['con', np.nan, np.nan, np.nan, np.nan]

harvest_factors_3 = harvest_factors_1.copy()
harvest_factors_3["value_2015"] = [0.3, 0.8, 1.0, 1.0, 2.0]

def test_correct_harvest_factors_input():
    """Test the function that checks harvest_factors input files,
    with correct input"""
    runner.silv.harvest.raw = harvest_factors_1
    runner.silv.harvest.extra_checks()

def test_incomplete_columns_returns_error():
    """Test the function that checks harvest_factors input files,
    with incomplete classifier column"""
    runner.silv.harvest.raw = harvest_factors_2
    with pytest.raises(ValueError) as excinfo:
        runner.silv.harvest.extra_checks()
    assert "cannot be incomplete" in str(excinfo.value)

def test_skew_factors_do_not_sum_to_one():
    """Test the function that checks harvest_factors input files
    with a skew factor that do not sum to one"""
    runner.silv.harvest.raw = harvest_factors_3
    with pytest.raises(ValueError) as excinfo:
        runner.silv.harvest.extra_checks()
    assert "do not sum to one" in str(excinfo.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.

- The core simulation tools are documented at `eu_cbm_hat.core`.
- Scenario combinations are defined in `eu_cbm_hat.combos` (the current
mechanism is subjected to change to allow user defined scenarios, provide
feedback under [issue
50](https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat/-/issues/50) ).
- The Harvest Allocation Tool is implemented in `eu_cbm_hat.cbm.dynamic`.

"""

# Special variables #
__version__ = '0.6.4'

# Built-in modules #
import os, sys

# First party modules #
from autopaths import Path
from autopaths.dir_path import DirectoryPath
from plumbing.git import GitRepo

# Constants #
project_name = 'eu_cbm_hat'
project_url  = 'https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat'

# Get paths to module #
self       = sys.modules[__name__]
module_dir = Path(os.path.dirname(self.__file__))

# The repository directory #
repos_dir = module_dir.directory

# The module is maybe in a git repository #
git_repo = GitRepo(repos_dir, empty=True)

# Where is the data, default case #
eu_cbm_data_dir = DirectoryPath("~/eu_cbm/eu_cbm_data/")

# But you can override that with an environment variable #
if os.environ.get("EU_CBM_DATA"):
    eu_cbm_data_dir = DirectoryPath(os.environ['EU_CBM_DATA'])

# Where are the AIDBs, default case
eu_cbm_aidb_dir = DirectoryPath("~/eu_cbm/eu_cbm_aidb/")

# But you can override that with an environment variable #
if os.environ.get("EU_CBM_AIDB"):
    eu_cbm_aidb_dir = DirectoryPath(os.environ['EU_CBM_AIDB'])



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 29 16:09:00 2022

@author: Viorel Blujdea and Paul Rougieux
"""

# Built-in modules #

# First party modules #
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.combos.base_combo import Combination
from eu_cbm_hat.cbm.dynamic       import DynamicRunner

class PIKFair(Combination):
    """
    A Combination used for the Harvest Allocation Tool (HAT).
    """

    short_name = 'pikfair'

    @property_cached
    def runners(self):
        """
        A dictionary of country codes as keys with a list of runners as
        values.
        """
        return {c.iso2_code: [DynamicRunner(self, c, 0)]
                for c in self.continent}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/deploy/eu_cbm_hat/scripts/setup/aidb_symlink.py
"""

# Built-in modules #
from eu_cbm_hat.core.continent import continent

# Create symlinks to every AIDB from every countries
for country in continent: country.aidb.symlink_all_aidb()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import re

# Third party modules #

# First party modules #

# Internal modules #

class InputYears:
    """
    This class will provide access to the years in all input files used by a runner

    Create a runner and list years used in its time dependent input data sets:

        >>> from eu_cbm_hat.core.continent import continent
        >>> r = continent.combos['special'].runners["LU"][-1]
        >>> input_years_dict = r.qaqc.input_years.dict
        >>> print(input_years_dict)
        >>> for key, value in input_years_dict.items():
        >>>     print(key, value, "\n")

    Display the max year for each data set:

        >>> {key: max(value) for key, value in input_years_dict.items()}

    Display the last year common to all input datasets:

        >>> print(r.qaqc.input_years.last_common())

    """

    def __init__(self, qaqc):
        # Default attributes #
        self.runner = qaqc.runner

    @property
    def dict(self):
        """Returns a dictionary with data set name as keys and
        lists of years as values.

        The output dictionary includes lists of year contained both
        in the input csv files
        and in the yaml definition of scenario combinations.
        """
        # Years in the harvest factor table
        harvest_cols = self.runner.silv.harvest.raw.columns
        harvest_years = (re.search(r"value_(\d+)", x) for x in harvest_cols)
        harvest_years = [int(m.group(1)) for m in harvest_years if m]

        # Years in the harvest tables
        irw_harvest_years = self.runner.harvest.irw.year.to_list()
        fw_harvest_years = self.runner.harvest.fw.year.to_list()

        # Years in the input events files for all activities combined
        timesteps = (self.runner.input_data["events"]
                     .sort_values(by=["step"])
                     .step.unique())
        events_years = self.runner.country.timestep_to_year(timesteps)

        # Place  the list of years in a dictionary
        dict1 = {"harvest_factor": harvest_years,
                 "irw_harvest": irw_harvest_years,
                 "fw_harvest": fw_harvest_years}

        # Load the lists of years used in the combo scenarios
        multi_year_input = ['events_templates',
                            'irw_frac_by_dist',
                            'harvest_factors',
                           ]
        combo_config = self.runner.combo.config
        dict2 = {"combo_" + m: list(combo_config[m].keys())
                 for m in multi_year_input}

        # Combine the two dictionaries
        dict1.update(dict2)
        return dict1

    def check_all_present(self):
        """Check that the time series are complete for each year in the input datasets.

        If a time series has a year missing, an error should be raised."""

    def last_common(self):
        """Returns the last common year available in all data sets
        """
        # Max value for each input data
        max_years = {key: max(value) for key, value in self.dict.items()}
        return  min(max_years.values())



"""The core components of the model are: `eu_cbm_hat.core.country`,
`eu_cbm_hat.core.continent` and `eu_cbm_hat.core.runner`. A runner is
associated to one country and runs the model for a specific combination of
input files called a scenario combination. The continent object contains a
dictionary of all scenario combinations documented at  `eu_cbm_hat.combos`. The
continent and scenario combination objects can be used together to create a
runner object for the test country `ZZ` as follows:

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['hat'].runners['ZZ'][-1]

"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# First party modules #
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.combos.base_combo import Combination
from eu_cbm_hat.core.runner import Runner

###############################################################################
class Historical(Combination):
    """
    This combo simulates the historical period ranging from
    the <country.start_year> until 2015, but no further.

    It uses the `reference` scenario for all activities.

    This combo represents a harvest that is pre-calculated and is not a
    function of the maximum wood supply. There is no interaction yet with the
    GFTM model.
    """

    short_name = 'historical'

    @property_cached
    def runners(self):
        """
        A dictionary of country codes as keys with a list of runners as
        values.
        """
        return {c.iso2_code: [HistoricalRunner(self, c, 0)]
                for c in self.continent}

###############################################################################
class HistoricalRunner(Runner):
    """
    Like a normal runner, but we redefine the num_timesteps property.
    """

    @property_cached
    def num_timesteps(self):
        """
        Compute the number of years we have to run the simulation for.
        Print all resulting years for each country:

            >>> from eu_cbm_hat.core.continent import continent
            >>> combo = continent.combos['historical']
            >>> for code, steps in combo.runners.items():
            >>>     r = steps[-1]
            >>>     print(code, ': ', r.num_timesteps)
        """
        # Retrieve parameters that are country specific #
        base_year      = self.country.base_year
        inv_start_year = self.country.inventory_start_year
        # Compute the number of years to simulate #
        period_max     = base_year - inv_start_year + 1
        # Return #
        return period_max



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to run the imaginary `ZZ` country to test the pipeline.
This version runs with the environment variable set to a temporary file.

Typically you would run this file from a command line like this:

    ipython3 -i -- ~/deploy/eu_cbm/eu_cbm_hat/scripts/running/run_zz_in_temp_dir_without_eu_cbm_data.py


On a development machine, you can update the ZZ dataset stored inside this repository with:

    ipython3 -i -- ~/deploy/eu_cbm/eu_cbm_hat/scripts/setup/copy_zz_from_libcbm_data.py

"""

from pathlib import Path
from tempfile import TemporaryDirectory
import os
import shutil

temp_dir = TemporaryDirectory()
dest_path = Path(temp_dir.name) / "eu_cbm_data"
# Define the environment variable
# This has to happen before we import anything from eu_cbm_hat
os.environ["EU_CBM_DATA"] = str(dest_path)

# Internal modules
from eu_cbm_hat import module_dir
orig_path = Path(module_dir) / "tests/eu_cbm_data"
# Copy ZZ test data to a temporary directory
shutil.copytree(orig_path, dest_path)

# This has to happen after we copy ZZ data to a temporary directory
from eu_cbm_hat.core.continent import continent
runner = continent.combos['hat'].runners['ZZ'][-1]
# Create the AIDB symlink
runner.country.aidb.symlink_all_aidb()
runner.num_timesteps = 30
# Run the test country ZZ
runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

# Remove the temporary directory
temp_dir.cleanup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A script to run Finland.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/repos/eu_cbm/eu_cbm_hat/scripts/running/run_fi.py

"""

from eu_cbm_hat.core.continent import continent
runner = continent.combos['reference'].runners['FI'][-1]
runner.country.base_year = 2021
runner.num_timesteps = 2050 - runner.country.inventory_start_year
runner.run(True, True, True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #

# First party modules #

# Internal modules #

###############################################################################
def make_classif_df(vals, clfrs):
    """
    Produces a dataframe useful for joining classifier values to
    other output dataframes. This dataframe looks like this:

              identifier timestep status forest_type region ...
        0              1        0    For          OB   LU00 ...
        1              2        0    For          OB   LU00 ...
        2              3        0    For          OB   LU00 ...
    """
    # Invert dictionary #
    vals = {v: k for m in vals.values() for k, v in m.items()}
    # Put actual values such as 'For' instead of numbers like '6' #
    clfrs = clfrs.set_index(['identifier', 'timestep'])
    clfrs = clfrs.replace(vals)
    clfrs = clfrs.reset_index()
    # Return #
    return clfrs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A script to run Slovenia

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/repos/eu_cbm/eu_cbm_hat/scripts/running/run_si.py

"""

from eu_cbm_hat.core.continent import continent


#############################################
# Declare which scenario combination to run #
#############################################
combo   = continent.combos['reference']
runner  = combo.runners['SI'][-1]
runner.num_timesteps = 50

# Run the model
output = runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from autopaths.auto_paths import AutoPaths

# Internal modules #

###############################################################################
class InputData:
    """
    This class will provide access to the input data of a Runner
    as several pandas data frames.
    The files listed here are the ones used to create the JSON that is
    consumed by `libcbm`.

    Example use:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['special'].runners["ZZ"][-1]
        >>> # Fetch the data from the country folder. 
        >>> # Necessary in case the runner has not been run yet.
        >>> runner.input_data()
        >>> print(runner.input_data["classifiers"])
        >>> print(runner.input_data["events"])

    """

    all_paths = """
    /input/csv/
    /input/csv/age_classes.csv         # Static
    /input/csv/classifiers.csv         # Static
    /input/csv/disturbance_types.csv   # Static
    /input/csv/events.csv              # Dynamic based on scenarios picked
    /input/csv/inventory.csv           # Dynamic based on scenarios picked
    /input/csv/transitions.csv         # Dynamic based on scenarios picked
    /input/csv/growth_curves.csv       # Dynamic based on scenarios picked
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        # Directories #
        self.paths = AutoPaths(self.parent.data_dir, self.all_paths)
        # Shortcuts #
        self.orig    = self.runner.country.orig_data
        self.combo   = self.runner.combo
        self.code    = self.runner.country.iso2_code
        self.act_dir = self.orig.paths.activities_dir

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    def __getitem__(self, item):
        return pandas.read_csv(str(self.paths[item]))

    #------------------------------- Methods ---------------------------------#
    def load(self, name):
        """Loads one of the dataframes."""
        # Load from CSV #
        df = self[name]
        # Return #
        return df

    def __call__(self, debug=False):
        """
        Create the input data files based on the scenario chosen for each
        different activity in the current combination.
        """
        # Message #
        self.parent.log.info("Preparing input data.")
        # Get the destination #
        csv_dir = self.paths.csv_dir
        csv_dir.remove()
        # The common static files just need to be copied over #
        common = self.orig.paths.common_dir
        common.copy(csv_dir)
        # Create the four dynamic files #
        for input_file in self.orig.files_to_be_generated:
            # The path to the file that we will create #
            out_path = self.paths[input_file]
            # What scenarios choices were made for this input file #
            choices = self.combo.config.get(input_file, {})
            # Initialize #
            result = pandas.DataFrame()
            # Optional debug message #
            msg = "Input file '%s' and combo '%s' for country '%s':"
            params = (input_file, self.combo.short_name, self.code)
            if debug: print(msg % params)
            # Iterate over every activity that is defined #
            for activity in choices:
                # Check it exists #
                if activity not in self.orig.activities:
                    msg = "The activity '%s' is not defined in '%s'."
                    raise FileNotFoundError(msg % (activity, self.act_dir))
                # Get the path to the file we will read #
                in_path = self.act_dir + activity + '/' + input_file + '.csv'
                # Read the file, but it's ok if it is empty or absent #
                try:
                    # We want to keep all values as objects and not floats #
                    df = pandas.read_csv(str(in_path), dtype=str)
                except (FileNotFoundError, pandas.errors.EmptyDataError):
                    continue
                # The scenario chosen for this activity and this input #
                scenario = choices[activity]
                # Filter rows to take only this scenario #
                df = df.query("scenario == '%s'" % scenario)
                # Optional debug message #
                msg = "   * for activity '%s', scenario '%s': %i rows"
                if debug: print(msg % (activity, scenario, len(df)))
                # Append #
                result = pandas.concat([result, df])
            # Remove the scenario column #
            if not result.empty: result = result.drop(columns=['scenario'])
            # Optional debug messages #
            if debug:
                print("   * result -> %i rows total\n" % len(result))
                print(result.dtypes[0:10], '\n-----------\n')
            # Write output #
            result.to_csv(str(out_path), index=False)
        # Return #
        return csv_dir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

from typing import Union, List

class PostProcessor(object):
    """
    This class will xxxx.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    def __call__(self):
        """
        xxxx.
        """
        return
        # Message #
        self.parent.log.info("Post-processing results.")
        # Lorem #
        pass

    def sum_flux_pool(self, by: Union[List[str], str], pools: List[str]):
        """Aggregate the flux pool table over the "by" variables and for the
        given list of pools.

        Example

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner_at = continent.combos["pikbau"].runners["AT"][-1]
            >>> living_biomass_pools = [
            >>>     "softwood_merch",
            >>>     "softwood_other",
            >>>     "softwood_foliage",
            >>>     "softwood_coarse_roots",
            >>>     "softwood_fine_roots",
            >>>     "hardwood_merch",
            >>>     "hardwood_foliage",
            >>>     "hardwood_other",
            >>>     "hardwood_coarse_roots",
            >>>     "hardwood_fine_roots",
            >>> ]
            >>> runner_at.post_processor.sum_flux_pool(by="year", pools=living_biomass_pools)
            >>> runner_at.post_processor.sum_flux_pool(by=["year", "forest_type"], pools=living_biomass_pools)

        """
        df = self.runner.output.pool_flux.groupby(by)[pools].sum()
        df.reset_index(inplace=True)
        return df



""" This script copies the ZZ data from the eu_cbm_data repository

It has to be run on a machine that has the eu_cbm_data repository. ZZ data is
treated in an identical manner to any other country and remains under version
control in the eu_cbm_data repository. However, for users who do not have
access to the private eu_cbm_data repository, we disseminate this data as part
of the eu_cbm_hat package. The purpose of this script is to copy data
necessary to run ZZ from the eu_cbm_data folder to a folder inside this package
eu_cbm_hat/tests/eu_cbm_data
"""

from pathlib import Path
import shutil
from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat import module_dir

# Path to copy data to
test_data_dir = Path(module_dir) / "tests/eu_cbm_data"
if not test_data_dir.exists():
    test_data_dir.mkdir(parents=True)

# File paths to copy data from
# After deleting all files not necessary to run ZZ
# from the eu_cbm_data repository
# I copied the output of the bash command
#     cd ~/rp/eu_cbm_data
#     find  combos/ common/ countries/ harvest/ -type f
files = """combos/reference.yaml
common/reference_years.csv
common/country_codes.csv
countries/ZZ/common/disturbance_types.csv
countries/ZZ/common/classifiers.csv
countries/ZZ/common/age_classes.csv
countries/ZZ/silv/harvest_factors.csv
countries/ZZ/silv/vol_to_mass_coefs.csv
countries/ZZ/silv/irw_frac_by_dist.csv
countries/ZZ/silv/events_templates.csv
countries/ZZ/activities/mgmt/inventory.csv
countries/ZZ/activities/mgmt/growth_curves.csv
countries/ZZ/activities/mgmt/transitions.csv
countries/ZZ/activities/mgmt/events.csv
countries/ZZ/activities/nd_nsr/inventory.csv
countries/ZZ/activities/nd_nsr/growth_curves.csv
countries/ZZ/activities/nd_nsr/transitions.csv
countries/ZZ/activities/nd_nsr/events.csv
countries/ZZ/activities/nd_sr/inventory.csv
countries/ZZ/activities/nd_sr/growth_curves.csv
countries/ZZ/activities/nd_sr/transitions.csv
countries/ZZ/activities/nd_sr/events.csv
countries/ZZ/activities/deforestation/inventory.csv
countries/ZZ/activities/deforestation/growth_curves.csv
countries/ZZ/activities/deforestation/transitions.csv
countries/ZZ/activities/deforestation/events.csv
countries/ZZ/activities/afforestation/inventory.csv
countries/ZZ/activities/afforestation/growth_curves.csv
countries/ZZ/activities/afforestation/transitions.csv
countries/ZZ/activities/afforestation/events.csv
countries/ZZ/config/associations.csv
domestic_harvest/reference/fw_harvest.csv
domestic_harvest/reference/irw_harvest.csv"""

for line in files.splitlines():
    orig_file = Path(eu_cbm_data_dir) / line
    dest_file = test_data_dir / line
    dest_file.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(orig_file, dest_file)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #

# Internal modules #
from eu_cbm_hat.pump.long_or_wide import events_wide_to_long

###############################################################################
class PreProcessor(object):
    """
    This class will update the input data of a runner based on a set of rules.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent  = parent
        self.runner  = parent
        self.country = parent.country
        # Shortcuts #
        self.input = self.runner.input_data

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    #--------------------------- Special Methods -----------------------------#
    def __call__(self):
        # Message #
        self.parent.log.info("Pre-processing input data.")
        # Check empty lines in all CSV inputs #
        for csv_path in self.all_csv: self.raise_empty_lines(csv_path)
        # Reshape the events file #
        self.reshape_events()
        # Check there are no negative timesteps #
        self.raise_bad_timestep()

    #----------------------------- Properties --------------------------------#
    @property
    def all_csv(self):
        """Get all CSV inputs in a list."""
        return [item.path_obj for item in self.input.paths._paths
                if item.path_obj.extension == 'csv']

    #------------------------------- Methods ---------------------------------#
    @staticmethod
    def raise_empty_lines(csv_path):
        """
        Loads one CSV files and raise an exception if there are any empty
        lines.
        """
        # Load from disk #
        try: df = pandas.read_csv(str(csv_path))
        # If the file is empty we can skip it #
        except pandas.errors.EmptyDataError: return
        # Get empty lines #
        empty_lines = df.isnull().all(1)
        # Check if there are any #
        if not any(empty_lines): return
        # Warn #
        msg = "The file '%s' has %i empty lines."
        raise Exception(msg % (csv_path, empty_lines.sum()))

    def reshape_events(self, debug=False):
        """Reshape the events file from the wide to the long format."""
        # The events file #
        path = self.input.paths.events
        # Optionally make a copy #
        if debug: path.copy(path.prefix_path + '_wide.csv')
        # Load it as a dataframe #
        wide = pandas.read_csv(str(path))
        # Reshape it #
        long = events_wide_to_long(self.country, wide)
        # Write to disk #
        long.to_csv(str(path), index=False)

    def raise_bad_timestep(self):
        """
        Raise an Exception if there are timesteps with a value below zero.
        """
        # Path to the file we want to check #
        path = str(self.input.paths.events)
        # Load from disk #
        try: df = pandas.read_csv(path)
        # If the file is empty we can skip it #
        except pandas.errors.EmptyDataError: return
        # Get negative values #
        negative_values = df['step'] < 0
        # Check if there are any #
        if not any(negative_values): return
        # Message #
        msg = "The file '%s' has %i negative values for the timestep column." \
              " This means you are attempting to apply disturbances to a" \
              " year that is anterior to the inventory start year configured."
        # Raise #
        raise Exception(msg % (path, negative_values.sum()))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import copy, math
import warnings

# Third party modules #
import pandas

# First party modules #
from plumbing.cache import property_cached
from libcbm.input.sit import sit_cbm_factory
from libcbm.model.cbm import cbm_variables

# Internal modules #
from eu_cbm_hat.cbm.simulation import Simulation
from eu_cbm_hat.core.runner import Runner
from eu_cbm_hat.info.silviculture import keep_clfrs_without_question_marks
# Constants #


class DynamicRunner(Runner):
    """
    Replaces the standard Simulation object with a DynamicSimulation instead.
    """

    @property_cached
    def simulation(self):
        """The object that can run `libcbm` simulations."""
        return DynamicSimulation(self)

###############################################################################
class DynamicSimulation(Simulation):
    """
    This class inherits from the standard Simulation object, and adds
    new functionality. It enables the running of dynamic simulations which
    can specify their disturbances just-in-time as the model is running.
    This is in contrast to standard simulations which must have all
    disturbances predefined before the model run.

    To see the simulation object:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['hat'].runners["ZZ"][-1]
        >>> runner.simulation.sources

    Compare to another scenario with skew by clear cut and thinning disturbances

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['skewfcth'].runners["ZZ"][-1]

    """

    # These are the dataframe (as attributes) returned by `cbm.step()` #
    df_names = ['classifiers', 'parameters', 'inventory',
                'state', 'flux', 'pools']

    # These are the source pools we want to track fluxes from #
    sources = ['softwood_merch',       'hardwood_merch',
               'softwood_other',       'hardwood_other',
               'softwood_stem_snag',   'hardwood_stem_snag',
               'softwood_branch_snag', 'hardwood_branch_snag']

    # These are the equivalent names in the libcbm dataframes #
    sources_cbm = ['SoftwoodMerch',      'HardwoodMerch',
                   'SoftwoodOther',      'HardwoodOther',
                   'SoftwoodStemSnag',   'HardwoodStemSnag',
                   'SoftwoodBranchSnag', 'HardwoodBranchSnag']

    #--------------------------- Special Methods -----------------------------#
    def dynamics_func(self, timestep, cbm_vars, debug=False):
        """
        First apply predetermined disturbances, then apply harvest
        specific to harvesting. The full specification for the "Harvest
        Allocation Tool" (H.A.T.) is described in:

            docs/harvest_allocation_specification.md

        Information used during development included:

        * The example notebook of the `libcbm` package.

            https://github.com/cat-cfs/libcbm_py/blob/master/examples/
            disturbance_iterations.ipynb
        """
        # Check if we want to switch the growth period classifier #
        if timestep == 1: cbm_vars = self.switch_period(cbm_vars)

        # Retrieve the current year #
        self.year = self.country.timestep_to_year(timestep)

        # Optional debug messages #
        if debug: print(timestep, self.year, self.country.base_year)

        # Run the usual rule based processor #
        cbm_vars = self.rule_based_proc.pre_dynamics_func(timestep, cbm_vars)

        # Check if we are still in the historical period #
        # If we are still in the historical period HAT doesn't apply
        if self.year < self.country.base_year:
            return cbm_vars

        msg = f"Time step {timestep} (year {self.year})."
        self.parent.log.info(msg)

        # Copy cbm_vars and hypothetically end the timestep here #
        end_vars = copy.deepcopy(cbm_vars)
        end_vars = cbm_variables.prepare(end_vars)
        end_vars = self.cbm.step(end_vars)

        # Check we always have the same sized dataframes #
        get_num_rows = lambda name: len(getattr(end_vars, name))
        assert len({get_num_rows(name) for name in self.df_names}) == 1

        # The age and land_class columns appears twice #
        renaming = {'age':        'inv_start_age',
                    'land_class': 'inv_start_land_class'}
        end_vars.inventory = end_vars.inventory.rename(columns=renaming)

        # Concatenate dataframes together by columns into one big df #
        stands = pandas.concat([getattr(end_vars, name)
                                for name in self.df_names], axis=1)

        # Check that the 'Input' column is always one and remove #
        assert all(stands['Input'] == 1.0)
        stands = stands.drop(columns='Input')

        # Get the columns that contain either pools or fluxes #
        cols = list(end_vars.flux.columns) + list(end_vars.pools.columns)
        cols.pop(cols.index('Input'))

        # Fluxes and pools are scaled to tonnes per one hectare so fix it #
        stands[cols] = stands[cols].multiply(stands['area'], axis="index")
#        stands[cols] = stands[cols].multiply(1000, axis="index")

        # Get the classifier columns #
        clfrs = list(self.country.orig_data.classif_names.values())

        # Get the classifier columns along with `disturbance_type` #
        cols = clfrs + ["disturbance_type"]

        # Get only eight interesting fluxes, summed also by dist_type #
        fluxes = stands.query("disturbance_type != 0")
        fluxes = fluxes.groupby(cols)
        fluxes = fluxes.agg({s + '_to_product': 'sum' for s in self.sources})
        fluxes = fluxes.reset_index()

        # Join the `irw` fractions with the fluxes going to `products` #
        irw_frac = self.runner.silv.irw_frac.get_year(self.year)
        clfrs_noq = keep_clfrs_without_question_marks(irw_frac, clfrs)
        cols_to_product = [s + "_to_product" for s in self.sources]
        sum_flux_before_merge = fluxes[cols_to_product].sum().sum()
        fluxes = fluxes.merge(irw_frac, how='left',
                              on=clfrs_noq + ["disturbance_type"],
                              suffixes=('_fluxes', ''))
        assert sum_flux_before_merge == fluxes[cols_to_product].sum().sum()
        missing_irw_frac = fluxes[self.sources].isna().any(axis=1)
        has_flux_to_prod =  fluxes[cols_to_product].sum(axis=1)>1
        selector = missing_irw_frac & has_flux_to_prod
        if any(selector):
            msg = "Industrial roundwood fractions defined in irw_frac_by_dist.csv "
            msg += "do not have irw fractions for the following classifiers:\n"
            msg += f"{fluxes[selector]}"
            raise ValueError(msg)

        # Join the wood density and bark fraction parameters also #
        coefs = self.runner.silv.coefs.df
        fluxes = fluxes.merge(coefs, how='left', on=['forest_type'])
        assert sum_flux_before_merge == fluxes[cols_to_product].sum().sum()

        # Calculate the total `flux_irw` and `flux_fw` for this year #
        def tot_flux_to_vol(irw=True):
            # Convert all fluxes' fraction to volume #
            tot = [fluxes[s + '_to_product'] *
                   (fluxes[s] if irw else (1 - fluxes[s])) *
                   (1 - fluxes['bark_frac']) /
                   (0.49 * fluxes['wood_density'])
                   for s in self.sources]
            # Sum to a scalar #
            return sum([s.sum() for s in tot])

        # The argument is False for firewood and True for roundwood #
        tot_flux_irw_vol = tot_flux_to_vol(irw=True)
        tot_flux_fw_vol  = tot_flux_to_vol(irw=False)

        # Fluxes to products from disturbances activities applied before HAT
        # salvage logging amount generated by HAT are *not* included here
        self.out_var('irw_predetermined', tot_flux_irw_vol)
        self.out_var('fw_predetermined',  tot_flux_fw_vol)

        # Get harvest for the current year #
        query  = "year == %s" % self.year
        harvest_irw_vol = self.runner.harvest.irw.query(query)['value']
        harvest_fw_vol  = self.runner.harvest.fw.query(query)['value']

        # Convert to a cubic meter float value #
        harvest_irw_vol = harvest_irw_vol.values[0] * 1000
        harvest_fw_vol  = harvest_fw_vol.values[0]  * 1000

        # add columns with harvests
        self.out_var('harvest_irw_vol', harvest_irw_vol)
        self.out_var('harvest_fw_vol',  harvest_fw_vol)

        # Calculate unsatisfied harvest #
        remain_irw_vol = harvest_irw_vol - tot_flux_irw_vol
        remain_fw_vol  = harvest_fw_vol  - tot_flux_fw_vol
        self.out_var('remain_irw_harvest', remain_irw_vol)
        self.out_var('remain_fw_harvest',  remain_fw_vol)

        # If there is no unsatisfied harvest, we stop here #
        if (remain_irw_vol <= 0) and (remain_fw_vol <= 0):
            return cbm_vars

        # To distribute remaining harvest, first load event templates #
        events = self.runner.silv.events.get_year(self.year)

        # Take only the stands that have not been disturbed yet #
        stands = stands.query("disturbance_type == 0")
        stands = stands.drop(columns = 'disturbance_type')

        # Keep only columns of interest from our current stands #
        interest = clfrs + ['time_since_last_disturbance',
                            'last_disturbance_type', 'age'] + self.sources_cbm
        stands = stands[interest]

        # Rename the pools to their snake case equivalent #
        stands = stands.rename(columns = dict(zip(self.sources_cbm,
                                                  self.sources)))
        # We will merge the current stands with the events templates #
        clfrs_noq = keep_clfrs_without_question_marks(events, clfrs)
        df = pandas.merge(stands, events, how='inner', on=clfrs_noq,
                          suffixes=('_stands', ''))

        # Convert last_disturbance_type from the libcbm stands to the id used in events_templates input
        dist_map = self.runner.simulation.sit.disturbance_id_map
        df["last_disturbance_id"] = df["last_disturbance_type"].map(dist_map)
        df["last_disturbance_id"] = df["last_disturbance_id"].astype(int)

        ################################
        # Filter eligible disturbances #
        ################################
        # We will filter on ages, `last_dist_id` and `min_since_last_dist` #
        df = df.query("age >= sw_start")
        df = df.query("age <= sw_end")
        df = df.query("last_dist_id == -1 | "
                      "last_dist_id == last_disturbance_id")
        df = df.query("min_since_last_dist == -1 | "
                      "min_since_last_dist <= time_since_last_disturbance")

        # If we have no stands to disturb, go straight to next year #
        if df.empty: return cbm_vars

        # We will now join the flux's proportions for each disturbance #
        props = self.runner.fluxes.df
        cols = self.runner.fluxes.cols + ['disturbance_type']
        df = pandas.merge(df, props[cols], how='left', on='disturbance_type')

        # Add the fractions going to `irw` and `fw` #
        mapping  = {pool: pool + '_irw_frac' for pool in self.sources}
        irw_frac = irw_frac.rename(columns = mapping)
        cols     = clfrs + ["disturbance_type"]
        df       = df.merge(irw_frac, how='left', on=cols)

        # Join the wood density and bark fraction parameters also #
        df = df.merge(coefs, how='left', on=['forest_type'])

        # Calculate the two volumes that would be produced by the events #
        def vol_by_source(row, source, irw):
            frac = row[source + '_irw_frac']
            frac = frac if irw else 1 - frac
            return (row[source] *
                    row[source + '_prod_prop'] *
                    frac *
                    (1 - row['bark_frac']) /
                    (0.49 * row['wood_density']))

        def mass_to_volume(row):
            irw_vol = (vol_by_source(row, s, True)  for s in self.sources)
            fw_vol  = (vol_by_source(row, s, False) for s in self.sources)
            return {'irw_vol': sum(irw_vol),
                    'fw_vol':  sum(fw_vol)}

        # Add two columns `irw_vol` and `fw_vol` to the dataframe #
        vols = df.apply(mass_to_volume, axis=1, result_type='expand')
        df = pandas.concat([df, vols], axis='columns')

        # Group our event candidates on classifiers and disturbance ID #
        grp_cols = cols + ['product_created']

        # All these columns must have unique values for a given age range #
        unique_cols = [col for col in events.columns if col not in grp_cols]
        unique_cols += ['wood_density', 'bark_frac']

        # Keep all required columns after the aggregation and sum volumes #
        agg_cols = {col: 'unique' for col in unique_cols}
        agg_cols['irw_vol'] = 'sum'
        agg_cols['fw_vol']  = 'sum'

        # Group-by and aggreagte so that age classes merge together #
        df = df.groupby(grp_cols)
        df = df.aggregate(agg_cols)
        df = df.reset_index()

        # Explode the uniques and check the number of rows does not change #
        orig_len = len(df)
        df = df.explode(unique_cols)
        assert len(df) == orig_len

        # Compute availability
        df['irw_avail'] = df['irw_vol'] / df['dist_interval_bias']
        df['fw_avail']  = df['fw_vol']  / df['dist_interval_bias']

        # If there is no extra industrial roundwood needed, set to zero #
        if remain_irw_vol <= 0.0:
            remain_irw_vol = 0.0
        else:
            if df['irw_vol'].sum() == 0.0:
                msg = "There is remaining IRW harvest this year, but there " \
                      "are no events that enable the creation of irw."
                raise Exception(msg)

        msg = f"IRW harvest {remain_irw_vol:.0f} m3."
        self.parent.log.info(msg)

        ######################
        # Harvest allocation #
        ######################
        # 1. Salvage logging disturbances
        #    - generate IRW and FW as a collateral
        #    - based on the normalized value of irw_avail
        # 2. Normal silviculture disturbances
        #    - generate IRW and FW as a collateral
        #    - based on the normalized value of irw_avail including a skew factor
        #    - the skew factor leads to a reshuffling of the harvest proportion by
        #      grouping variables such as by coniferous and broadleaves
        #      (but this is flexible as it depends only on the columns that are
        #      filled in the harvest_factors.csv file)
        # 3. Fuel wood only disturbances
        #    - generate FW only
        #    - based on the normalized value of irw_avail
        # Separate the disturbances data frame between
        # salvage logging, `irw_and_fw` and `fw_only`
        salv = (df["last_dist_id"] != -1) & (df["product_created"] == "irw_and_fw")
        silv = (df["last_dist_id"] == -1) & (df["product_created"] == "irw_and_fw")
        fw_only = (df["product_created"] == "fw_only")
        df_irw_salv = df.loc[salv].copy()
        df_irw_silv = df.loc[silv].copy()
        df_fw  = df.loc[fw_only].copy()

        # Check that all rows are covered once and only once by the sub data frames
        alloc_check = (salv.astype(int) + silv + fw_only) == 1
        if not all(alloc_check):
            msg = "Some disturbances are present in more than one category"
            msg += f"\n{df.loc[alloc_check]}"
            raise Exception(msg)
        assert all(salv | silv | fw_only)
        assert(len(df) == len(df_irw_salv) + len(df_irw_silv) + len(df_fw))

        # Check `products_created` is correct and not lying #
        check_irw = df_irw_silv.query("fw_vol == 0.0")
        check_fw  = df_fw.query("irw_vol != 0.0")
        if not check_irw.empty:
            cols = ["forest_type", "disturbance_type", "product_created"]
            cols += ["irw_avail", "fw_avail"]
            msg = "Some rows have zero values for the fuel wood column.\n"
            msg += f"{check_irw[cols]}"
            raise ValueError(msg)
        assert check_fw.empty

        # Process salvage logging disturbances in priority if they are present
        if any(salv):
            # irw and fw potential from salvage logging disturbances
            irw_salv_avail = df_irw_salv["irw_avail"].sum()
            fw_salv_avail = df_irw_salv["fw_avail"].sum()
            # Print a message #
            msg = "Potential amount available from salvage logging: "
            msg += f"{irw_salv_avail:.0f} m3 IRW and "
            msg += f"{fw_salv_avail:.0f} m3 fw (colateral of IRW)."
            self.parent.log.info(msg)

            # If the harvest is greater than the potential, allocate only the potential
            irw_to_allocate = min(irw_salv_avail, remain_irw_vol)

            # Distribute evenly according to the potential irw volume produced
            # compute the proportion only for the salvage logging disturbances
            df_irw_salv["irw_norm"] = (df_irw_salv["irw_avail"] /
                                            df_irw_salv["irw_avail"].sum())

            # Calculate how much volume we need from each stand #
            df_irw_salv['irw_need'] = irw_to_allocate * df_irw_salv['irw_norm']
            assert math.isclose(df_irw_salv["irw_need"].sum(),
                                irw_to_allocate)
        else:
            irw_salv_avail = 0
            fw_salv_avail = 0

        # Save salvage logging in output
        self.out_var('irw_salv_avail', irw_salv_avail)
        self.out_var('fw_salv_avail', fw_salv_avail)

        # If salvage logging didn't satisfies all harvest
        # Continue allocating disturbances
        if irw_salv_avail < remain_irw_vol:
            remain_irw_vol_after_salv = remain_irw_vol - irw_salv_avail
            msg = f"Remaining IRW harvest after salvage logging {remain_irw_vol_after_salv:.0f} m3.\n"
            self.parent.log.info(msg)
            # Distribute evenly according to the potential irw volume produced #
            df_irw_silv["irw_norm"] = (df_irw_silv["irw_avail"] /
                                       df_irw_silv["irw_avail"].sum())

            # Skew the normalized value based on the harvest skew factors
            # We will retrieve the harvest skew factors for the current year #
            harvest = self.runner.silv.harvest.get_year(self.year)

            # Only one of the columns matches the current year #
            harvest = harvest.rename(columns = {'value_%i' % self.year: 'skew'})

            # Keep only IRW coefficients
            harvest = harvest[harvest["product_created"] == "irw_and_fw"]


            # Keep only the columns that are not empty as join columns
            harvest_join_cols = []
            for col in self.runner.silv.harvest.cols:
                if not any(harvest[col].isna()):
                    harvest_join_cols.append(col)
            harvest = harvest[harvest_join_cols + ['skew']]

            # Check if all rows for which a skew factor is defined are really
            # present in df_irw_silv
            df_irw_silv_check = df_irw_silv.value_counts(harvest_join_cols).reset_index()
            if len(df_irw_silv_check) < len(harvest):
                msg += "Some skew factors are present in harvest but not present in "
                msg += "df_irw_silv. This might happen in rare cases where there is only "
                msg += "coniferous forest available and no broadleaf forest available. "
                msg += "For example because broadleaves are too young and the "
                msg += "query('age >= sw_start') excluded those broadleaf rows. "
                self.parent.log.info(msg)
                # Recompute the skew
                harvest2 = df_irw_silv_check.merge(harvest, on=harvest_join_cols)
                harvest2["skew"] = harvest2["skew"] / harvest2["skew"].sum()
                harvest = harvest2

            # If silv_practice is defined in harvest factors then use
            # self.runner.fluxes.df to add the silv_practice column to
            # df_irw_silv
            if "silv_practice" in harvest_join_cols:
                df_silv_practice = self.runner.fluxes.df[["disturbance_type", "silv_practice"]]
                df_irw_silv = df_irw_silv.merge(df_silv_practice, on="disturbance_type")

                # Check silv practices are actually present in disturbance_types.csv
                hfsp = harvest.silv_practice.unique()
                distsp = self.runner.fluxes.df["silv_practice"].unique()
                if not set(hfsp).issubset(distsp):
                    msg = "silv_practice defined in harvest_factors.csv: "
                    msg += f"{hfsp}"
                    msg +=  " do not match the ones in disturbance_types.csv: "
                    msg += f"{distsp}"
                    raise ValueError(msg)

            # Aggregate the normalized value by groups
            df_irw_silv["irw_norm_agg"] = df_irw_silv.groupby(harvest_join_cols)["irw_norm"].transform(sum)

            # Merge disturbances and harvest factors
            df_irw_silv = pandas.merge(df_irw_silv,
                                       harvest,
                                       how='inner',
                                       on=harvest_join_cols)

            # Modify the harvest distribution coefficient by the skew along each grouping variable
            df_irw_silv["irw_norm_skew"] = (df_irw_silv["irw_norm"]
                                            * df_irw_silv["skew"]
                                            / df_irw_silv["irw_norm_agg"])

            # Raise an error if irw_norm_skew does not sum to one
            if not math.isclose(df_irw_silv["irw_norm_skew"].sum(), 1):
                msg = "IRW norm skew doesn't sum to one:"
                msg += f"{df_irw_silv.groupby(harvest_join_cols)['irw_norm_skew'].sum()}\n"
                msg += "The normalized available merchantable roundwood is distributed as follows:\n"
                msg += f"{df_irw_silv.groupby(harvest_join_cols)['irw_norm_agg'].unique()}\n"
                msg += "The harvest factors are distributed as follows:\n"
                msg += f"{ harvest[harvest_join_cols + ['skew']]}\n "
                msg += "This means that some combinations of silvicultural practices "
                msg += "are not present in the events template."
                msg += "Correct the input in havest_factors.csv."
                msg += "It could also be that the share of industrial roundwood is not specified."
                msg += "Check the irw_frac_by_dist.csv"
                raise ValueError(msg)

            potential_irw = df_irw_silv["irw_avail"].sum()
            msg += f"Potential IRW amount available from remaining disturbances: "
            msg += f"{potential_irw:.0f} m3 IRW."
            prct = 100 * remain_irw_vol / potential_irw
            msg += f"\nIRW harvest corresponds to {prct:.0f}% of the annualized available potential."
            self.parent.log.info(msg)

            #df_irw_silv["prop"] = df_irw_silv[

            # Calculate how much volume we need from each stand #
            df_irw_silv["irw_need"] = (remain_irw_vol_after_salv *
                                             df_irw_silv["irw_norm_skew"])
            assert math.isclose(df_irw_silv["irw_need"].sum(), remain_irw_vol_after_salv)
            # The user is free to over allocate IRW, but will be a warning if the
            # allocation is over the potential annualized availability.
            if  remain_irw_vol > potential_irw:
                excess_prct = remain_irw_vol / potential_irw - 1
                excess_prct = round(excess_prct*100)
                msg = f"\nIRW harvest is greater than the annualized potential by {excess_prct}%."
                warnings.warn(msg)

        # Combine IRW disturbance from salvage logging with normal silviculture operations
        df_irw = pandas.concat([df_irw_salv, df_irw_silv])

        # Create columns if they were not created above i.e.
        # in case there was no IRW harvest at all
        if not "irw_need" in df_irw.columns:
            df_irw["irw_need"] = 0
            df_irw["irw_norm"] = 0

        # Check again whether the irw amount is fully allocated
        assert math.isclose(df_irw['irw_need'].sum(), remain_irw_vol)

        # Check the collateral fuel wood generated
        # How much is this volume as compared to the total volume possible #
        df_irw['irw_frac'] = df_irw['irw_need'] / df_irw['irw_vol']

        # How much firewood would this give us as a collateral product #
        df_irw['fw_colat'] = df_irw['irw_frac'] * df_irw['fw_vol']

        # Subtract from remaining firewood harvest #
        still_remain_fw_vol = remain_fw_vol - df_irw['fw_colat'].sum()
        self.out_var('still_remain_fw_vol', still_remain_fw_vol)
        colat_prct = (df_irw['fw_colat'].sum() / remain_fw_vol) * 100
        msg = f"FW Demand (remaining after FW salvage) {remain_fw_vol:.0f} m3 . "
        msg += f"Collateral FW (generated by IRW dist) {df_irw['fw_colat'].sum():.0f} m3 "
        msg += f"i.e. the ratio FW demand / colat FW is {colat_prct:.0f}%."
        self.parent.log.info(msg)

        # If there is no extra firewood needed, set to zero #
        if still_remain_fw_vol <= 0.0:
            still_remain_fw_vol = 0.0
        else:
            if df_fw['fw_vol'].sum() == 0.0:
                msg = "There is remaining fw harvest this year:"
                msg += f"{round(still_remain_fw_vol)} m3, "
                msg += "but there are no events that enable the creation of fw only."
                raise Exception(msg)

        # If there is still firewood to satisfy, distribute it evenly
        # Note: in case `still_remain_fw_vol` is equal to zero,
        # the events amount equal to zero will be filtered out later
        df_fw['fw_norm'] = df_fw['fw_avail'] / df_fw['fw_avail'].sum()
        df_fw['fw_need'] = still_remain_fw_vol * df_fw['fw_norm']
        assert math.isclose(df_fw['fw_need'].sum(), still_remain_fw_vol)

        potential_fw = df_fw['fw_avail'].sum()
        if potential_fw == 0:
            msg = f"Annualized potential FW disturbances is {potential_fw} "
            msg += f"and still remaining potential is {still_remain_fw_vol}."
        else:
            harvest_pot_percent = still_remain_fw_vol / potential_fw
            msg = f"Still remaining fuel wood harvest represents {harvest_pot_percent*100:.0f}% "
            msg += "of the annualized potential FW disturbances."
        self.parent.log.info(msg)

        # The user is free to over allocate fw, but will be a warning if the
        # allocation is over the potential annualized availability.
        if  still_remain_fw_vol > potential_fw:
            excess_prct = still_remain_fw_vol / potential_fw - 1
            excess_prct = round(excess_prct*100)
            msg = "\nStill remaining fuel wood harvest is greater than "
            msg += "the annualized potential FW disturbances."
            msg += f"by {excess_prct}%. "
            warnings.warn(msg)

        # Convert to mass (we don't need to care about source pools) #
        df_irw['amount'] = ((df_irw['irw_need'] + df_irw['fw_colat']) *
                            (0.49 * df_irw['wood_density']) /
                            (1 - df_irw['bark_frac']))
        df_fw['amount']  = (df_fw['fw_need'] *
                            (0.49 * df_fw['wood_density']) /
                            (1 - df_fw['bark_frac']))

        # Put the two dataframes back together #
        df = pandas.concat([df_irw, df_fw])

        # Filter out any events that have an amount of zero #
        df = df.query("amount != 0.0").copy()

        # Convert IDs back from the SIT standard to the user standard #
        df = self.conv_dists(df)
        df = self.conv_clfrs(df)


        df.insert(0, 'year', self.year)
        cols = ['year'] +  clfrs

        # Prepare the remaining missing columns for the events #
        df['measurement_type'] = 'M'
        df['step'] = timestep
        df = df.rename(columns={'disturbance_type': 'dist_type_name'})

        # Save some columns of this dataframe as a CSV in the output
        self.out_var('tot_irw_vol_avail', df['irw_avail'].sum())
        self.out_var('tot_fw_vol_avail',  df['fw_avail'].sum())

        # Select which events columns appear in the output record
        # Note: dist_type_name is already an input dist id (not an internal dist id)
        # It will not be converted by the outputdata.__setitem__() method
        cols += ['dist_type_name', 'product_created', 'dist_interval_bias',
                 'using_id', 'sw_start', 'sw_end', 'hw_start', 'hw_end',
                 'min_since_last_dist', 'max_since_last_dist', 'last_dist_id',
                 'sort_type', 'measurement_type', 'efficiency', 'skew', 'wood_density',
                 'bark_frac', 'irw_avail', 'fw_avail',
                 'irw_norm', 'irw_need', 'irw_frac',
                 'fw_colat', 'fw_norm', 'fw_need', 'amount']
        # Write the events to an output file for the record
        self.runner.output.events = pandas.concat([self.runner.output.events, df[cols]])

        # Get only the right columns in the dataframe to send to `libcbm` #
        cols = self.runner.input_data['events'].columns
        df = df[cols].copy()

        # Create disturbances and send the events to libcbm
        dyn_proc = sit_cbm_factory.create_sit_rule_based_processor(
            self.sit,
            self.cbm,
            reset_parameters = False,
            sit_events = df
        )

        # Run the dynamic rule based processor #
        cbm_vars = dyn_proc.pre_dynamics_func(timestep, cbm_vars)

        # Print a message #
        msg = f"Time step {timestep} (year {self.year}) is about to finish.\n"
        self.parent.log.info(msg)

        # Return #
        return cbm_vars

    #--------------------------- Other Methods -------------------------------#
    def conv_dists(self, df):
        """
        Convert the disturbance IDs from their internal simulation IDs that
        are defined by SIT into the user defined equivalent string.
        """
        # Get the conversion mapping #
        id_to_id = self.runner.simulation.sit.disturbance_id_map
        # Apply the mapping to the dataframe #
        df['disturbance_type'] = df['disturbance_type'].map(id_to_id)
        # Return #
        return df

    def conv_clfrs(self, df):
        """
        Convert the classifier IDs from their internal simulation IDs that
        are defined by SIT into the user defined equivalent string.
        """
        # Get all the conversion mappings, for each classifier #
        all_maps = self.runner.simulation.sit.classifier_value_ids.items()
        # Apply each of them to the dataframe #
        for classif_name, str_to_id in all_maps:
            # Keep question marks as is
            if df[classif_name].unique()[0] == "?":
                continue
            # Convert other values
            mapping = {v:k for k,v in str_to_id.items()}
            df[classif_name] = df[classif_name].map(mapping)
        # Return #
        return df

    def out_var(self, key, value):
        self.runner.output.extras.loc[self.year, key] = value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

This script will combine the two files:

* /export/yields.csv
* /export/historical_yields.csv

into one single file.

The word "yield" will be dropped because it is a reserved keyword in python and
replaced by "growth".

* /export/growth_curves.csv

This is useful to switch all the yield curves of every species as soon as we
hit the country start year.
"""

# Built-in modules #

# Third party modules #
from tqdm import tqdm

# First party modules #
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.core.continent import continent as libcbm_continent

# Continents #
from cbmcfs3_runner.core.continent import continent as cbmcfs3_continent

###############################################################################
class MergeGrowthCurves(object):

    def __init__(self, cbmcfs3_country):
        # Main attributes #
        self.cbmcfs3_country = cbmcfs3_country

    @property_cached
    def libcbm_country(self):
        """The matching libcbm country object."""
        return libcbm_continent.countries[self.cbmcfs3_country.iso2_code]

    def __call__(self, verbose=False):
        # Get paths #
        hist_curves = self.cbmcfs3_country.orig_data.paths.yields
        curr_curves = self.cbmcfs3_country.orig_data.paths.historical_yields
        destination = self.libcbm_country.orig_data.paths.growth_curves
        # Print output #
        if verbose:
            print(hist_curves)
            print(curr_curves)
            print(destination)
            print('----')
        # Combine #
        hist_curves.copy(destination)
        lines = iter(curr_curves)
        next(lines)
        destination.writelines(lines, mode='a')
        # Check headers are the same #
        assert hist_curves.first == curr_curves.first == destination.first
        # Remove old file #
        old_file = self.libcbm_country.data_dir + '/orig/csv/yield.csv'
        old_file.remove()

###############################################################################
if __name__ == '__main__':
    # Make converter objects, one per country #
    mergers = [MergeGrowthCurves(c) for c in cbmcfs3_continent]
    # Run them all #
    for merger in tqdm(mergers):
        merger()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import simplejson as json

# First party modules #

###############################################################################
class CreateJSON(object):
    """
    This class will generate the JSON file consumed by `libcbm`.

    The mapping mode "SeparateAdminEcoClassifiers" is used by
    libcbm/input/sit/sit_mapping.py to assign the spatial unit IDs.

    We use the same keys as defined in the dictionary below, except
    for `yield` being translated to `growth_curves` since `yield` is a
    reserved keyword in python.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    def __call__(self):
        self.runner.paths.json.write(json.dumps(self.content,
                                                indent     = 4,
                                                ignore_nan = True))

    #-------------------------- Class attributes -----------------------------#
    template = {
        "import_config": {
            "age_classes":       {"type": "csv", "params": {"path": None}},
            "classifiers":       {"type": "csv", "params": {"path": None}},
            "disturbance_types": {"type": "csv", "params": {"path": None}},
            "events":            {"type": "csv", "params": {"path": None}},
            "inventory":         {"type": "csv", "params": {"path": None}},
            "transitions":       {"type": "csv", "params": {"path": None}},
            "yield":             {"type": "csv", "params": {"path": None}},
          },
        "mapping_config": {
            "spatial_units": {
                "mapping_mode":     "SeparateAdminEcoClassifiers",
                "admin_classifier": "region",
                "eco_classifier":   "climate",
                "admin_mapping":    None,
                "eco_mapping":      None,
            },
            "disturbance_types": None,
            "species": {
                "species_classifier": "forest_type",
                "species_mapping":    None,
            }
        }
    }

    #----------------------------- Properties --------------------------------#
    @property
    def content(self):
        # Make a copy of the template #
        config = self.template.copy()
        # Get the mapping config sub dictionary #
        maps = config['mapping_config']
        # Retrieve the four classifiers mappings as a dict #
        mappings = self.runner.country.associations.all_mappings
        # Set the admin and eco classifiers #
        maps['spatial_units']['admin_mapping'] = mappings['map_admin_bound']
        maps['spatial_units']['eco_mapping']   = mappings['map_eco_bound']
        # Set the disturbances and species classifiers #
        maps['disturbance_types']              = mappings['map_disturbance']
        maps['species']['species_mapping']     = mappings['map_species']
        # Get the names of all the CSVs #
        all_files_names = config['import_config'].keys()
        # Finally set the paths to all the CSVs #
        for file_name in all_files_names:
            # Special case for the `yields` file #
            name = file_name if file_name != 'yield' else 'growth_curves'
            # Get the path #
            full_path = self.runner.input_data.paths[name]
            # Set it in the dictionary #
            config['import_config'][file_name]['params']['path'] = full_path
        # Return result #
        return config


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

This script loads the harvest that originates from the economic model
called "GFTMX". This economic model is not run dymamically, instead it
was run once in the past and the outputs recorded in the `eu_cbm_data`
repository.

The wood product harvest used in the `eu_cbm_hat` simulations is seen as the
production from the perspective of the economic model.

Usage:

    >>> from eu_cbm_hat.info.harvest import gftm_outputs
    >>> from eu_cbm_hat.info.harvest import combined

Related issues:

* https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat/-/issues/8

This script provides a dictionary `gftm_outputs` that looks like this:

    'reference':   {'irw': <OutputGFTM object>,
                    'fw':  <OutputGFTM object>},
    'market_drop': {'irw': <OutputGFTM object>,
                    'fw':  <OutputGFTM object>}, ...}

It also provides a `combined` dictionary.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat.core.country import all_country_codes

# Constant directory for all the data #
harvest_dir = eu_cbm_data_dir + 'domestic_harvest/'

# Constant file names for every scenario #
roundwood = "irw_harvest.csv"
fuelwood  = "fw_harvest.csv"

# A mapping between country names and codes #
country_to_iso2 = all_country_codes[['country', 'iso2_code']]

###############################################################################
def wide_to_long(df):
    """Convert a harvest dataset to the long format."""
    # List columns for pivot #
    cols = ['faostat_name', 'element', 'unit', 'country']
    # Convert #
    df = pandas.wide_to_long(df,
                             stubnames = 'value',
                             i         = cols,
                             j         = 'year',
                             sep       = '_')
    # Reset index #
    df = df.reset_index()
    # Return #
    return df

###############################################################################
class OutputGFTM:
    """
    Access to the raw harvest coming from the GFTM model, for a particular
    scenario and a particular wood type. We also provide a filtered and
    transformed version of the data.
    """

    def __init__(self, scenario, csv_name):
        self.scenario = scenario
        self.csv_name = csv_name

    @property_cached
    def csv_path(self):
        return harvest_dir + self.scenario + '/' + self.csv_name

    @property
    def raw(self):
        return pandas.read_csv(self.csv_path)

    @property_cached
    def df(self):
        # Load #
        df = self.raw
        # Wide to long transform #
        df = wide_to_long(df)
        # Add country codes (and remove all unknown countries) #
        df = df.merge(country_to_iso2, on='country')
        # Remove other columns that give no information #
        df = df[['iso2_code', 'year', 'value']]
        # Return #
        return df

###############################################################################
class Demand:
    """
    Access the specific harvest values for the current simulation run.
    For both industrial roundwood and fuelwood.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        # Shortcuts #
        self.combo = self.runner.combo
        self.code  = self.runner.country.iso2_code
        # Choices made for harvest in the current combo #
        self.choices = self.combo.config['harvest']

    def make_df(self, wood_type):
        # Case number 1: there is only a single scenario specified #
        if isinstance(self.choices, str):
            df = gftm_outputs[self.choices][wood_type].df
        # Case number 2: the scenarios picked vary according to the year #
        else:
            df = self.choices.merge(combined[wood_type],
                                    how = 'left',
                                    on = ['year', 'scenario'])
        # Filter for current country #
        df = df.query("iso2_code == '%s'" % self.code)
        # Check there is data left #
        assert not df.empty
        # Return #
        return df

    #----------------------------- Properties --------------------------------#
    @property_cached
    def irw(self): return self.make_df('irw')
    @property_cached
    def fw(self):  return self.make_df('fw')

###############################################################################
# Figure out all possible harvest scenarios that we have #
scenarios = [d.name for d in harvest_dir.flat_directories]

# Load all the GFTM outputs for every scenario and every wood type #
gftm_outputs = {scen: {'irw': OutputGFTM(scen, roundwood),
                       'fw':  OutputGFTM(scen, fuelwood)}
                for scen in scenarios}

# A function to combine all scenarios together for a given wood type #
def make_combined(wood_type):
    df = [out[wood_type].df for out in gftm_outputs.values()]
    df = pandas.concat(df, keys=scenarios, names=['scenario'])
    df = df.reset_index(level='scenario').reset_index(drop=True)
    return df

# Provide two dataframes that contain all scenarios combined #
combined = {'irw': make_combined('irw'),
            'fw':  make_combined('fw')}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# First party modules #
from autopaths.auto_paths import AutoPaths
from plumbing.cache       import property_cached
from plumbing.logger      import create_file_logger
from plumbing.timer       import LogTimer

# Internal modules #
from eu_cbm_hat.cbm.simulation      import Simulation
from eu_cbm_hat.info.harvest         import Demand
from eu_cbm_hat.info.fluxes         import Fluxes
from eu_cbm_hat.info.input_data     import InputData
from eu_cbm_hat.info.internal_data  import InternalData
from eu_cbm_hat.info.output_data    import OutputData
from eu_cbm_hat.info.silviculture   import Silviculture
from eu_cbm_hat.launch.create_json  import CreateJSON
from eu_cbm_hat.pump.post_processor import PostProcessor
from eu_cbm_hat.pump.pre_processor  import PreProcessor
from eu_cbm_hat.qaqc                import Qaqc
import eu_cbm_hat

# Third party modules

###############################################################################
class Runner(object):
    """
    This object is capable of running a CBM simulation pipeline, starting
    from a few input tables, such as an inventory and a list of disturbances
    and to bring this data all the way to the predicted carbon stock and
    fluxes.

    You can run a combo like this:

        >>> from eu_cbm_hat.core.continent import continent
        >>> combo  = continent.combos['historical']
        >>> runner = combo.runners['LU'][0]
        >>> runner.run()

    The runner has an attribute `output` that only deals with final output
    data that can be reached after closing and reopening your interpreter.

    The runner has an attribute `internal` that only deals with getting the
    data from the libcbm objects as they are while still in RAM and not
    written to disk. Examples:

    >>> runner.internal.classif_df
    If you are running a simluation and want to get the information from RAM.

    >>> runner.output.classif_df
    If you are analyzing results that were run in the past and want to get the
    information from disk.
    """

    all_paths = """
    /input/
    /input/json/config.json
    /output/
    /logs/runner.log
    """

    def __init__(self, combo, country, num):
        # Base attributes #
        self.combo   = combo
        self.country = country
        self.num     = num
        # How to reference this runner #
        self.short_name  = self.combo.short_name + '/'
        self.short_name += self.country.iso2_code + '/'
        self.short_name += str(self.num)
        # Where the data will be stored for this run #
        self.data_dir = self.combo.output_dir + self.short_name + '/'
        # Automatically access paths based on a string of many subpaths #
        self.paths = AutoPaths(self.data_dir, self.all_paths)

    def __repr__(self):
        return '%s object on "%s"' % (self.__class__, self.data_dir)

    def __bool__(self): return self.paths.log.exists

    #---------------------------- Compositions -------------------------------#
    @property_cached
    def create_json(self):
        return CreateJSON(self)

    @property_cached
    def simulation(self):
        """The object that can run `libcbm` simulations."""
        return Simulation(self)

    @property_cached
    def pre_processor(self):
        """Update the input data to this run using some rules."""
        return PreProcessor(self)

    @property_cached
    def post_processor(self):
        """Update or convert the output data to this run using some rules."""
        return PostProcessor(self)

    @property_cached
    def input_data(self):
        """
        Access the input data to this run. This data can be
        a modified version of the original country's CSV files.
        """
        return InputData(self)

    @property_cached
    def qaqc(self):
        """
        Quality Assurance and Quality Control on the runner input and output data.
        """
        return Qaqc(self)

    @property_cached
    def harvest(self):
        """Access the specific harvest values for this simulation run."""
        return Demand(self)

    @property_cached
    def silv(self):
        """Access to silvicultural data."""
        return Silviculture(self)

    @property_cached
    def fluxes(self):
        """
        Returns a dataframe linking disturbance to proportions sent to
        the pool `products`.
        """
        return Fluxes(self)

    @property_cached
    def output(self):
        """Create and access the output data to this run."""
        return OutputData(self)

    @property_cached
    def internal(self):
        """
        Access and format data concerning the simulation as it is being
        run.
        """
        return InternalData(self)

    #----------------------------- Properties --------------------------------#
    @property_cached
    def log(self):
        """
        Each runner will have its own logger. By default we clear the log file
        when we start logging. This happens when you call this property for
        the first time. If you want to view the log file of a previous run,
        check the attribute `self.paths.log`.
        """
        # Pick console level #
        level = 'error'
        if hasattr(self, 'verbose'):
            if isinstance(self.verbose, bool):
                if self.verbose:
                    level = 'debug'
            else: level = self.verbose
        # Create #
        logger = create_file_logger(self.short_name,
                                    self.paths.log,
                                    console_level = level)
        # Return #
        return logger

    @property
    def tail(self):
        """A short summary showing just the end of the log file."""
        msg  = "\n## Runner `%s`\n" % self.short_name
        msg += "\nTail of the log file at `%s`\n" % self.paths.log
        msg += self.paths.log.pretty_tail
        return msg

    @property_cached
    def num_timesteps(self):
        """
        Compute the default number of years we have to run the simulation for.
        To do this, we select the disturbance with the highest time step.
        """
        # Load #
        df = self.input_data.load('events')
        # Compute #
        period_max = df['step'].max()
        # Return #
        return period_max

    #------------------------------- Methods ---------------------------------#
    def run(self, keep_in_ram=False, verbose=True, interrupt_on_error=False):
        """
        Run the full modelling pipeline for a given country, a given combo
        and a given step.
        """
        # Verbosity level #
        self.verbose = verbose
        # Messages #
        self.log.info("Using %s." % eu_cbm_hat)
        self.log.info("Runner '%s' starting." % self.short_name)
        # Start the timer #
        self.timer = LogTimer(self.log)
        self.timer.print_start()
        # Clean everything from previous run #
        self.remove_directories()
        # Create the input data #
        self.input_data()
        # Check silviculture input tables
        self.silv.check()
        # Modify input data, combos can subclass this #
        self.modify_input()
        # Pre-processing #
        self.pre_processor()
        # Create the JSON configuration #
        self.create_json()
        # Run the model #
        self.timer.print_elapsed()
        self.simulation(interrupt_on_error)
        self.timer.print_elapsed()
        # Save the results to disk #
        if self.simulation.error is not True: self.output.save()
        # Free memory #
        if not keep_in_ram: self.simulation.clear()
        # Post-processing #
        self.post_processor()
        # Messages #
        self.timer.print_end()
        self.timer.print_total_elapsed()
        # Final message #
        if self.simulation.error is not True: msg = "Done."
        else: msg = "Done with errors."
        self.log.info(msg)
        # Return #
        return self.output

    def remove_directories(self):
        """
        Removes the directory that will be recreated by running this runner.
        This guarantees that all output data is regenerated.
        Note: we need to keep the log we are writing to currently.
        """
        # Message #
        self.log.info("Removing directory '%s'." % self.data_dir.with_tilda)
        # The output directory #
        self.paths.input_dir.remove(safe=False)
        self.paths.output_dir.remove(safe=False)
        # Empty all the other logs found there except ours #
        for element in self.paths.logs_dir.flat_contents:
            if element != self.paths.log:
                element.remove()

    #--------------------------- Special Methods -----------------------------#
    def modify_input(self):
        """Combos can subclass this at will."""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.

"""

# Built-in modules #
import textwrap

# Third party modules #
import yaml, pandas
from p_tqdm import p_umap, t_map

# First party modules #
from autopaths      import Path
from plumbing.cache import property_cached
from plumbing.timer import Timer

# Internal modules #
from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat.core.runner import Runner

# Constant directory for all the data #
yaml_dir = eu_cbm_data_dir + 'combos/'

###############################################################################
class Combination(object):
    """
    This object represents a combination of specific scenarios for different
    activities and includes any other customization of a given model run.

    Each Combination subclass must define a list of Runner instances as
    the <self.runners> property. This enables the complete customization of
    any Runner by the specific Combination instance.

    You can run a combo like this:

        >>> from eu_cbm_hat.core.continent import continent
        >>> combo = continent.combos['historical']
        >>> combo()

    You can run a specific runner from a given country like this:

        >>> from eu_cbm_hat.core.continent import continent
        >>> combo = continent.combos['historical']
        >>> r = combo.runners['LU'][-1]
        >>> r.run(True, True, True)

    You can then check the output pools:

        >>> r.output.load('pools')
    """

    short_name = None

    def __init__(self, continent):
        # Save parent #
        self.continent = continent
        # The combos dir used for all output #
        self.output_dir = self.continent.output_dir
        # The base dir for our output #
        self.base_dir = Path(self.output_dir + self.short_name + '/')

    def __repr__(self):
        return '%s object with %i runners' % (self.__class__, len(self))

    def __iter__(self): return iter(self.runners.values())
    def __len__(self):  return len(self.runners.values())

    def __getitem__(self, key):
        """Return a runner based on a country code."""
        return self.runners[key]

    #----------------------------- Properties --------------------------------#
    @property_cached
    def config(self):
        """
        The values chosen by the user in the YAML file which decide on every
        scenario choice for every activity and silvicultural practice.
        """
        # The path to our specific YAML file #
        yaml_path = yaml_dir + self.short_name + '.yaml'
        # Read it with a third party library #
        with open(yaml_path, "r") as handle:
            result = yaml.safe_load(handle)
        # Convert silvicultural choices to dataframes #
        key = 'harvest'
        value = result[key]
        if not isinstance(value, str):
            df = pandas.DataFrame.from_dict(value,
                                            orient  = 'index',
                                            columns = ['scenario'])
            df = df.rename_axis('year').reset_index()
            result[key] = df
        # Return result #
        return result

    @property_cached
    def runners(self):
        """
        A dictionary of country codes as keys with a list of runners as
        values.
        """
        return {c.iso2_code: [Runner(self, c, 0)] for c in self.continent}

    #------------------------------- Methods ---------------------------------#
    def __call__(self, parallel=False, timer=True):
        """A method to run a combo by simulating all countries."""
        # Message #
        print("Running combo '%s'." % self.short_name)
        # Timer start #
        timer = Timer()
        timer.print_start()
        # Function to run a single country #
        def run_country(args):
            code, steps = args
            for runner in steps:
                return runner.run()
        # Run countries sequentially #
        if not parallel:
            result = t_map(run_country, self.runners.items())
        # Run countries in parallel #
        else:
            result = p_umap(run_country, self.runners.items(), num_cpus=4)
        # Timer end #
        timer.print_end()
        timer.print_total_elapsed()
        # Compile logs #
        self.compile_logs()
        # Return #
        return result

    def compile_logs(self, step=-1):
        # Open file #
        summary = self.base_dir + 'all_logs.md'
        summary.open(mode='w')
        # Write title #
        title = "# Summary of all log files #\n\n"
        summary.handle.write(title)
        # Loop over runners #
        for rs in self.runners.values():
            r = rs[step]
            summary.handle.write("\n## " + r.country.country_name)
            summary.handle.write(' (' + r.country.iso2_code + ')' + '\n\n')
            content = textwrap.indent(r.paths.log.contents, '    ')
            summary.handle.write(content)
        # Close #
        summary.close()
        # Message #
        msg = "Log files compiled at:\n\n%s\n"
        print(msg % summary)
        # Return #
        return summary



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

This script will convert the AIDB from the old MS Access format supported by
CBM-CFS3 to the new SQLite format supported by libcbm.

Typically you would run this file from a command line like this on a windows
 system which has repositories in the home directory:

     ipython3.exe -i -- %HOMEPATH%/repos/eu_cbm_hat/scripts/conversion/aidb.py

You need to run this on a machine that has a Microsoft Access driver installed.
So likely this will require a Windows machine.

If you try on a Unix machine, even with `unixodbc` installed you will get an
error such as:

    Error: ('01000', "[01000] [unixODBC][Driver Manager]
    Can't open lib 'Microsoft Access Driver (*.mdb, *.accdb)'
    : file not found (0) (SQLDriverConnect)")

If you want to run only one country for testing, you can place
this line in the main loop:

        if converter.cbmcfs3_country.iso2_code != 'ZZ': continue
"""

# Built-in modules #
import os

# Third party modules #
from tqdm import tqdm

# First party modules #
import cbm_defaults.app
from plumbing.cache import property_cached
from plumbing.timer import Timer

# Internal modules #
from eu_cbm_hat.core.continent import continent as libcbm_continent

# Continents #
from cbmcfs3_runner.core.continent import continent as cbmcfs3_continent

# Constants #
home = os.path.expanduser("~")
ddl_path = home + r"\repos\cbm_defaults\schema\cbmDefaults.ddl"

###############################################################################
class ConvertAIDB(object):
    """
    This class will enable us to convert the old AIDB format (Access DB)
    from CBMCFS3 to an SQLite3 database compatible with `libcbm_py`.
    We will use only one locale, and keep "en-CA".
    """

    template = {
        "output_path": None,
        "schema_path": ddl_path,
        "default_locale": "en-CA",
        "locales": [
            {
             "id":    1,
             "code": "en-CA"
             }
        ],
        "archive_index_data": [
            {
             "locale": "en-CA",
             "path":   None
             }
        ]
    }

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__,
                                        self.cbmcfs3_country.iso2_code)

    def __init__(self, cbmcfs3_country):
        # Main attributes #
        self.cbmcfs3_country = cbmcfs3_country

    @property_cached
    def libcbm_country(self):
        """The matching libcbm country object."""
        return libcbm_continent.countries[self.cbmcfs3_country.iso2_code]

    def __call__(self, verbose=True, aidb_repo=True):
        # Get the source path #
        source = self.cbmcfs3_country.paths.aidb_eu
        # Get the destination path #
        if aidb_repo: destin = self.libcbm_country.aidb.repo_file
        else:         destin = self.libcbm_country.aidb.paths.db
        # Messages #
        if verbose:
            print("\nCountry: " + self.cbmcfs3_country.iso2_code)
            print("Source: " + source)
            print("Destination: " + destin)
            print("-------------")
        # Make a copy of the template dictionary #
        config = self.template.copy()
        # Specify paths in this dictionary #
        config['archive_index_data'][0]['path'] = source
        config['output_path'] = destin
        # Delete previous version of the database #
        destin.remove()
        # Call method in `cbm_defaults` #
        cbm_defaults.app.run(config)

###############################################################################
if __name__ == '__main__':
    # Make converter objects, one per country #
    converters = [ConvertAIDB(c) for c in cbmcfs3_continent]
    # Show which version of the repository we are using #
    print("Using %s." % cbm_defaults.app)
    print("-------------")
    # Print timer start #
    timer = Timer()
    timer.print_start()
    # Run them all #
    for converter in tqdm(converters):
        converter()
    # Print end #
    timer.print_end()
    timer.print_total_elapsed()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import numpy

# First party modules #

# Internal modules #

def check_question_marks(df, df_name, classifiers):
    """Raise an error if classifiers have a mixture of
    question mark and no question mark"""
    for classif_name in classifiers:
        values = df[classif_name].unique().tolist()
        if len(values) > 1 and "?" in values:
            msg =  "Mixture of question marks and other values "
            msg += f"not allowed in {df_name}.\n"
            msg += f"The unique values of the {classif_name} column are: {values}"
            raise ValueError(msg)

class SilvCheck:
    """
    Check the consistency of silviculture input files

    Whenever possible tests are based on raw input files available before the run
    has started, when SIT is not available yet.

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['special'].runners["ZZ"][-1]

    Check that fuel wood disturbances don't generate industrial roundwood:

        >>> runner.qaqc.silv_check.fw_doesnt_create_irw()

    The check is based on the `events_templates`and `irw_frac_by_dist`

        >>> runner.silv.events.raw
        >>> runner.silv.irw_frac.raw

    List disturbance ids used in the input data "activities" directory and in
    the "silv" events_templates.csv.

        >>> # Fetch the data from the country folder. This call is only
        >>> # necessary in case the runner has not been run yet.
        >>> runner.input_data()
        >>> runner.qaqc.silv_check.dist_ids_activities()
        >>> runner.qaqc.silv_check.dist_ids_silv_events_templates()

    Check classifiers with question marks

        >>> runner.qaqc.silv_check.check_classifier_with_question_marks()

    """
    def __init__(self, qaqc):
        # Default attributes #
        self.runner = qaqc.runner
        # Disturbance mapping
        assoc_df = self.runner.country.associations.df
        self.assoc = assoc_df.loc[assoc_df["category"] == "MapDisturbanceType"]

    def fw_doesnt_create_irw(self):
        """Check that fuel wood only disturbances don't generate industrial roundwood"""
        index = ["disturbance_type", "product_created"]
        prod_by_dist = (self.runner.silv.events.raw
                        .set_index(index).index
                        .unique().to_frame(index=False))
        fw_by_dist = prod_by_dist[prod_by_dist["product_created"] == "fw_only"]

        # Keep industrial roundwood fractions only for the fuel wood disturbances
        df = (self.runner.silv.irw_frac.raw
              .merge(fw_by_dist, on="disturbance_type", how="inner"))

        # Exclude identifier columns to find the name of the value columns
        cols = self.runner.silv.irw_frac.raw
        identifiers =  self.runner.silv.irw_frac.dup_cols
        identifiers += ["dist_type_name"]
        val_cols = list(set(cols) - set(identifiers))
        # assert value columns are zero
        agg_cols = {col: "sum" for col in val_cols}
        df_agg = df.groupby(["disturbance_type", "product_created"])
        df_agg = df_agg.agg(agg_cols)
        if not numpy.allclose(df_agg.sum(), 0):
            msg = "fuel wood only disturbances "
            msg += "should not generate industrial roundwood:\n"
            msg += f"{df_agg}"
            raise ValueError(msg)

    def dist_ids_activities(self):
        """List disturbance ids used in the input data "activities" folder"""
        df = self.runner.input_data["events"]
        df = df.value_counts("dist_type_name")
        df = df.reset_index(name="number_of_rows")
        return df

    def dist_ids_silv_events_templates(self):
        """List disturbance ids used in the "silv" events_templates.csv"""
        df = self.runner.silv.events.raw
        df = df.value_counts(["disturbance_type", "dist_type_name"])
        df = df.reset_index(name="number_of_rows")
        return df

    def check_classifier_with_question_marks(self):
        """Check the input tables for classifiers with question marks"""
        clfrs = list(self.runner.country.orig_data.classif_names.values())
        irw_frac = self.runner.country.orig_data.load("irw_frac")
        check_question_marks(irw_frac, "irw_frac", clfrs)
        # Do not check the input events
        # input_events = self.runner.input_data["events"]
        # check_question_marks(input_events, "input events from the activities folder", clfrs)

    def cc_th_and_dist_matrix_proportion(self):
        """Check that clear cut disturbances are the ones that remove most of the merch biomass

        Rank fluxes to merchantable pool to check clear cut disturbances
        are actually the ones with the highest proportion in the dist matrix

        ! requires SIT

        Example use :

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['hat'].runners['ZZ'][-1]
            >>> # Run at least one present time step so that SIT is available
            >>> runner.num_timesteps = runner.country.base_year - runner.country.inventory_start_year + 1
            >>> runner.run()
            >>> runner.qaqc.silv_check.cc_th_and_dist_matrix_proportion()

        """
        cols = ['user_name', 'silv_practice', 'hardwood_merch_prod_prop', 'softwood_merch_prod_prop']
        df = self.runner.fluxes.df.sort_values("hardwood_merch_prod_prop", ascending=False)[cols]
        return df

    def check_cc_th_and_dist_matrix_proportion(self, threshold=0.79):
        merch_cols = ['hardwood_merch_prod_prop', 'softwood_merch_prod_prop']
        df = self.runner.cc_th_and_dist_matrix_proportion()
        assert all(df.loc[df["silv_practice"]=="th"][merch_cols] < threshold)
        assert all(df.loc[df["silv_practice"]=="cc"][merch_cols] > threshold)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #

# First party modules #
from autopaths.auto_paths import AutoPaths
from plumbing.cache       import property_cached

# Internal modules #
from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat.core.country import Country
from eu_cbm_hat.combos       import combo_classes

###############################################################################
class Continent(object):
    """
    Entry object to the pipeline.

    Aggregates countries together and enables access to a data frame containing
    concatenated data from all countries at once.
    """

    all_paths = """
    /countries/
    /output/
    """

    def __init__(self, base_dir):
        """
        Store the directory paths where there is a directory for every
        country and for every combo.
        """
        # The base directory #
        self.base_dir = base_dir
        # Automatically access paths based on a string of many subpaths #
        self.paths = AutoPaths(eu_cbm_data_dir, self.all_paths)
        # Where the input data will be stored #
        self.countries_dir = self.paths.countries_dir
        # Where the output data will be stored #
        self.output_dir = self.paths.output_dir

    def __repr__(self):
        return '%s object with %i countries' % (self.__class__, len(self))

    def __getitem__(self, key):
        """Return a runner based on a tuple of combo, country and step."""
        return self.get_runner(*key)

    def __iter__(self): return iter(self.countries.values())
    def __len__(self):  return len(self.countries.values())

    #----------------------------- Properties --------------------------------#
    @property_cached
    def countries(self):
        """Return a dictionary of country iso2 codes to country objects."""
        all_countries = [Country(self, d)
                         for d in self.countries_dir.flat_directories]
        return {c.iso2_code: c for c in all_countries}

    @property_cached
    def combos(self):
        """Return a dictionary of combination names to Combination objects."""
        all_combos = [combo(self) for combo in combo_classes]
        return {s.short_name: s for s in all_combos}

    #------------------------------- Methods ---------------------------------#
    def get_runner(self, combo, country, step):
        """Return a runner based on combo, country and step."""
        return self.combos[combo].runners[country][step]

###############################################################################
# Create singleton #
continent = Continent(eu_cbm_data_dir)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import os

# Third party modules #
from tqdm import tqdm
import pandas

# First party modules #
from autopaths.auto_paths import AutoPaths

# Internal modules #
from eu_cbm_hat                    import eu_cbm_data_dir
from eu_cbm_hat.core.continent     import continent
from eu_cbm_hat.pump.long_or_wide  import events_long_to_wide
from eu_cbm_hat.pump.pre_processor import PreProcessor

# Constants #
interface_dir = eu_cbm_data_dir + 'interface/'

###############################################################################
class MakeActivities(object):
    """
    This class will change the structure of the input data directory for
    each country, and set it up for the new features containing both
    activities, scenarios and combinations of the latter.

    More information is contained in the notebooks of the `bioeconomy_notes`
    repository. An example directory structure is the following:

        ZZ
        ├── activities
        │   ├── afforestation
        │   │   ├── events.csv
        │   │   ├── growth_curves.csv
        │   │   ├── inventory.csv
        │   │   └── transitions.csv
        │   ├── deforestation
        │   │   └── events.csv
        │   ├── mgmt
        │   │   ├── events.csv
        │   │   ├── growth_curves.csv
        │   │   ├── inventory.csv
        │   │   └── transitions.csv
        │   ├── nd_nsr
        │   │   ├── events.csv
        │   │   └── transitions.csv
        │   └── nd_sr
        │       ├── events_wide.csv
        │       └── transitions.csv
        ├── common
        │   ├── age_classes.csv
        │   ├── classifiers.csv
        │   └── disturbance_types.csv
        ├── orig
        │   ├── aidb.db -> eu_cbm_aidb/countries/ZZ/orig/config/aidb.db
        │   └── associations.csv
        └── silv
            ├── product_types.csv
            └── silvicultural_practices.csv
    """

    #------------------------------ File lists -------------------------------#
    common_list = ['disturbance_types.csv', 'classifiers.csv',
                   'age_classes.csv']

    silv_list = ['product_types.csv', 'silvicultural_practices.csv']

    config_list = ['associations.csv', 'aidb.db']

    mgmt_list = ['events.csv', 'inventory.csv', 'transitions.csv',
                 'growth_curves.csv']

    activities = ['afforestation', 'deforestation', 'mgmt', 'nd_nsr', 'nd_sr']

    #------------------------------ Autopaths --------------------------------#
    old_all_paths = """
    /orig/
    /orig/csv/
    /orig/csv/age_classes.csv        
    /orig/csv/classifiers.csv        
    /orig/csv/disturbance_types.csv  
    /orig/csv/events.csv             
    /orig/csv/inventory.csv          
    /orig/csv/transitions.csv        
    /orig/csv/growth_curves.csv      
    /orig/config/
    /orig/config/associations.csv        
    /orig/config/aidb.db       
    """

    new_all_paths = """
    /common/
    /common/age_classes.csv         
    /common/classifiers.csv         
    /common/disturbance_types.csv   
    /activities/         
    /activities/mgmt/events.csv           
    /activities/mgmt/inventory.csv        
    /activities/mgmt/transitions.csv      
    /activities/mgmt/growth_curves.csv  
    /config/
    /config/associations.csv
    /config/aidb.db       
    /silv/  
    /silv/product_types.csv
    /silv/silvicultural_practices.csv
    """

    #--------------------------- Special Methods -----------------------------#
    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.country.iso2_code)

    def __init__(self, country):
        # Default attributes #
        self.country = country
        # AutoPaths #
        self.old_paths = AutoPaths(self.country.data_dir, self.old_all_paths)
        self.new_paths = AutoPaths(self.country.data_dir, self.new_all_paths)

    def __call__(self):
        # Move existing files
        self.move_stuff()
        # Create empty files for all possible activities #
        self.create_activities()
        # Switch events files to the wide format #
        self.make_events_wide()
        # Add the scenario column to every file #
        self.add_scen_column()
        # Fix the transitions file #
        self.restore_header()

    #------------------------------- Methods ---------------------------------#
    def move_stuff(self):
        # Common #
        for item in self.common_list:
            self.old_paths[item].move_to(self.new_paths[item])
        # Silv #
        for item in self.silv_list:
            self.new_paths[item].touch()
        # Config #
        for item in self.config_list:
            self.old_paths[item].move_to(self.new_paths[item])
        # Mgmt #
        for item in self.mgmt_list:
            self.old_paths[item].move_to(self.new_paths[item])
        # Remove old directories #
        self.country.data_dir.remove_empty_dirs()

    def create_activities(self):
        # Other activities #
        for act in self.activities:
            if act == 'mgmt': continue
            for item in self.mgmt_list:
                directory = self.new_paths.activities_dir + act + '/'
                directory.create_if_not_exists()
                file = directory + item
                file.touch()

    def make_events_wide(self):
        # The path to the mgmt events file #
        path = self.new_paths.events
        # Read it #
        long = pandas.read_csv(str(path))
        # Get a pre-processor #
        pre_proc = PreProcessor(type('X', (), {'country': self.country}))
        # Transform it #
        wide = events_long_to_wide(self.country, long)
        # Write it #
        wide.to_csv(str(path), index=False)
        # Return #
        return str(path)

    def add_scen_column(self):
        # LU was already done previously #
        if self.country.iso2_code == 'LU': return
        # The files #
        files_to_be_modify = ['growth_curves',
                              'transitions',
                              'inventory']
        # Create the four dynamic files #
        for input_file in files_to_be_modify:
            # The path to the file that we will modify #
            path = self.new_paths[input_file]
            # Read the file #
            df = pandas.read_csv(str(path))
            # Add column #
            df.insert(0, 'scenario', 'reference')
            # Write output #
            df.to_csv(str(path), index=False, float_format='%g')

    def restore_header(self):
        """
        In a pandas dataframe, the column names have to be unique, because
        they are implemented as an index. However in the file
        "transition_rules.csv", column names are repeated.
        So we have to restore these headers afterwards.
        """
        # Read from disk #
        header = self.new_paths.transitions.first
        # Modify #
        header = header.split(',')
        header = [n.replace('.1', '') for n in header]
        header = ','.join(header)
        # Write to disk #
        self.new_paths.transitions.remove_first_line()
        self.new_paths.transitions.prepend(header)

    #------------------------- The flat symlinks -----------------------------#
    @property
    def country_interface_dir(self):
        return interface_dir + self.country.iso2_code + '/'

    @property
    def interface_base(self):
        return  self.country_interface_dir + self.country.iso2_code + '_'

    def make_interface(self, hardlinks=True, debug=False):
        """
        This method can create symlinks to the input files in a flat hierarchy,
        in essence providing a user interface to the input data.

        This was originally developed to be compatible with Excel. The Excel
        software has the ridiculous limitation of not being able to open two
        files with the same name.

        Moreover, in the case of windows, symbolic links don't overcome this
        limitation and Excel still complains when it opens two symbolic links
        that point to different files.

        This issue is not fixed by using hard links instead of symbolic links.
        This is because Excel never modifies a given file. When saving, it
        creates a temporary file in the same directory, then deletes the
        original file and renames the temporary file to the name of the
        original file. This destroys the hard links upon every save operation.
        """
        # Create the directory #
        self.country_interface_dir.create_if_not_exists()
        # Same case for all of: "Common, Silv, Config" #
        for item in self.common_list + self.silv_list + self.config_list:
            file = self.new_paths[item]
            dest = self.interface_base + 'config_' + file.name
            dest.remove()
            if debug: print(str(file), " -> ", str(dest))
            if hardlinks: os.link(str(file), str(dest))
            else:                 file.link_to(dest)
        # Different case for "Activities" #
        for subdir in self.new_paths.activities_dir.flat_directories:
            act = subdir.name
            for file in subdir.flat_files:
                dest = self.interface_base + act + '_' + file.name
                dest.remove()
                if debug: print(str(file), " -> ", str(dest))
                if hardlinks: os.link(str(file), str(dest))
                else:                 file.link_to(dest)
        # Return #
        return self.interface_base

    #------------------------ Copying files back -----------------------------#
    def save_interface(self, debug=False):
        """
        In the end, the only way to make this `interface` work is to have a
        script copy every file in the flat hierarchy back to it's expected
        place within the `eu_cbm_data` repository.
        """
        # Same case for all of: "Common, Silv, Config" #
        for item in self.common_list + self.silv_list + self.config_list:
            file = self.new_paths[item]
            source = self.interface_base + 'config_' + file.name
            if debug: print(str(source), " -> ", str(file))
            source.copy_to(file)
        # Different case for "Activities" #
        for subdir in self.new_paths.activities_dir.flat_directories:
            act = subdir.name
            for file in subdir.flat_files:
                source = self.interface_base + act + '_' + file.name
                if debug: print(str(source), " -> ", str(file))
                source.copy_to(file)
        # Return #
        return self.interface_base

###############################################################################
makers = [MakeActivities(c) for c in continent]
if __name__ == '__main__': print([maker() for maker in tqdm(makers)])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

Usage:

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['special'].runners["ZZ"][-1]
    >>> runner.qaqc.aidb_check.run_all_checks()

"""

# Built-in modules #

# Third party modules #
import numpy

# First party modules #

# Internal modules #

class AIDBCheck:
    """
    Check the consistency of the AIDB

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['special'].runners["ZZ"][-1]

    Check id duplication in vol_to_bio_factor

        >>> runner.qaqc.aidb_check.check_vol_to_bio_factor_id_duplication()

    """
    def __init__(self, qaqc):
        # Default attributes #
        self.runner = qaqc.runner
        self.aidb = self.runner.country.aidb

    def run_all_checks(self):
        """Run all AIDB checks"""
        self.check_vol_to_bio_factor_id_duplication()

    def check_vol_to_bio_factor_id_duplication(self):
        """Investigate identifier duplication in the vol_to_bio_factor in the AIDB"""
        df = self.aidb.db.read_df('vol_to_bio_factor')
        # Select duplicated rows
        selector = df["id"].duplicated(keep=False)
        if any(selector):
            msg = "Duplicated ids in the vol_to_bio_factor table in the AIDB."
            msg += "The following rows are duplicated:\n"
            msg += f"{df[selector]}\n\n"
            msg += "Investigate by loading the table with a command similar to this one:\n"
            msg += "runner.country.aidb.db.read_df('vol_to_bio_factor')"
            raise ValueError(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A script to run Sweden.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/deploy/eu_cbm/eu_cbm_hat/scripts/running/run_se.py
"""

from eu_cbm_hat.core.continent import continent
runner = continent.combos['reference'].runners['SE'][-1]
runner.num_timesteps = 50

# Run the model
runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

# Check expected provided
runner.qaqc.expected_provided.by(index=["forest_type", "disturbance_type"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to run Austria.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/deploy/eu_cbm_hat/scripts/running/run_at.py

"""

from eu_cbm_hat.core.continent import continent
runner  = continent.combos['reference'].runners['AT'][-1]
runner.num_timesteps = 25

# Run the model
output = runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

# Input events sent to libcbm
events_input = runner.input_data["events"]
# Events stored in the output
events_output = runner.output["events"]
output_extras = runner.output["extras"]







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from autopaths.dir_path   import DirectoryPath
from plumbing.cache       import property_cached

# Internal modules #
from eu_cbm_hat                     import eu_cbm_data_dir
from eu_cbm_hat.launch.associations import Associations
from eu_cbm_hat.info.orig_data      import OrigData
from eu_cbm_hat.info.aidb           import AIDB

# Country codes #
all_country_codes = eu_cbm_data_dir + 'common/country_codes.csv'
all_country_codes = pandas.read_csv(str(all_country_codes))

# Country reference years #
ref_years = eu_cbm_data_dir + 'common/reference_years.csv'
ref_years = pandas.read_csv(str(ref_years))

###############################################################################
class Country(object):
    """
    This object gives access to the data pertaining to one country
    amongst the 26 EU member states we are examining.
    """

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.iso2_code)

    def __init__(self, continent, data_dir=None):
        """Store the data directory paths where everything will start from."""
        # Parent #
        self.continent = continent
        # Main directory #
        self.data_dir = DirectoryPath(data_dir)
        # Set country codes #
        self.set_codes()
        # Store the reference years #
        self.set_years()

    #---------------------------- Compositions -------------------------------#
    @property_cached
    def associations(self):
        """
        Associations of admin/eco/species/disturbances names between
        the input and the reference.
        """
        return Associations(self)

    @property_cached
    def aidb(self):
        """Archive Index Database also called `cbm_defaults` in libcbm."""
        return AIDB(self)

    @property_cached
    def orig_data(self):
        """Access to the immutable original data."""
        return OrigData(self)

    @property_cached
    def cbmcfs3_country(self):
        """Access the corresponding country in the `cbmcfs3_runner` module."""
        from cbmcfs3_runner.core.continent import continent as cbm_continent
        return cbm_continent.countries[self.iso2_code]

    #----------------------------- Properties --------------------------------#
    @property_cached
    def combos(self):
        """
        A dictionary linking combo names to a list of runners
        that concern only this country.
        """
        from cbmcfs3_runner.core.continent import continent
        return {n: s.runners[self.iso2_code]
                for n, s in continent.combos.items()}

    #------------------------------- Methods ---------------------------------#
    def set_codes(self):
        """
        Update all the country codes for this country.
        Typically the result will look something like this:

         'iso2_code':      'BE',
         'country_num':     255,
         'country_name':   'Belgium',
         'country_m49':     56,
         'country_iso3':   'BEL',
         'nuts_zero_2006': 'BE',
         'nuts_zero_2016': 'BE',
        """
        # The reference ISO2 code #
        self.iso2_code = self.data_dir.name
        # Load name mappings #
        selector = all_country_codes['iso2_code'] == self.iso2_code
        # Check that we know about this country #
        if not selector.any():
            msg = "The directory '%s' is not a country that is known."
            raise ValueError(msg % self.data_dir)
        # Get the right row #
        row = all_country_codes.loc[selector].iloc[0]
        # Store all the country references codes #
        self.country_num  = row['country_code']
        self.country_name = row['country']
        self.country_m49  = row['m49_code']
        self.country_iso3 = row['iso3_code']
        # More crazy codes #
        self.nuts_zero_2006 = row['nuts_zero_2006']
        self.nuts_zero_2016 = row['nuts_zero_2010']

    def set_years(self):
        """Update all the reference years for this country."""
        # This is the same for all countries #
        self.base_year = 2021
        # This is different for each country.
        # The `inventory_start_year` is the oldest year in the inventory data
        # reported by the national forest inventory.
        row = ref_years.loc[ref_years['country'] == self.iso2_code].iloc[0]
        self.inventory_start_year = row['inv_start_year']

    def timestep_to_year(self, timestep):
        """
        Time step 0 is the output of the makelist (so called "spin-up")
        procedure. It represents the initial state.

        Will convert a Series containing simulation time-steps such as:
           [1, 2, 3, 4, 5]
        to actual corresponding simulation years such as:
           [1990, 1991, 1992, 1993, 1994]
        """
        return timestep + self.inventory_start_year

    def year_to_timestep(self, input_year):
        """
        The reverse operation of `timestep_to_year`.

        Will convert a Series containing years such as:
           [1990, 1991, 1992, 1993, 1994]
        to simulation time-steps such as:
           [1, 2, 3, 4, 5]

        Checking the consistency between the 2 functions:

            >>> from eu_cbm_hat.core.continent import continent
            >>> country = continent.countries['ZZ']
            >>> year = country.timestep_to_year([1,2])
            >>> country.year_to_timestep(year)
            array([1, 2])
        """
        return input_year - self.inventory_start_year



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.

A list of all combo classes.

Creating a new scenario combination:


- In `eu_cbm_data`

    - Add data specific to the “new_scenario” in each relevant input file. The
    assumptions have to be identified as “new_scenario” on the column
    “scenario” across all files. 

    - For example, in case of a new scenario for afforestation, specific
    records must be added in: events.csv, growth_curves.csv, inventory.csv,
    transitions.csv. in

        ...\eu_cbm_data\countries\ZZ\activities\afforestation.

    - Also, information specific to the new assumptions must be added in the
    following files contained in the other directory (as explained above in
    Lower level of the EU-CBM-HAT: country specific inputs):

        ...\eu_cbm_data\countries\ZZ\common - on the new disturbance
        types;
        ...\eu_cbm_data\countries\ZZ\config - for the changes in the AIDB
        and association file;
        ...\eu_cbm_data\countries\ZZ\silv – description of the silvicultural
        practices and wood use.

    - Add data on IRW and FW demands in the new directory. irw_demand.csv,
    fw_demands.csv, rw_demands.csv in the directory
    ...\eu_cbm_data\demand\new_scenario.

    - In `eu_cbm_data/combos` Compile a new yaml file with the combination of
    scenarios desired. NB. A yaml file is a configuration file that can be
    edited with a text editor.

- In `eu_cbm_hat`

    - Create a new file in ...\eu_cbm_hat\combos: new_scenario.py. Create a new
    class and import the corresponding module in the file __init__.py in
    ...\eu_cbm_hat\combos of the eu_cbm_hat Add the new class
    [...new_scenario...] 

    - and import the new module [...new_...] in __init__.py, by paying
    attention to consistency of names, e.g., class name: “New_scenarios”,
    short_name: “new_scenarios”.

"""

# Built-in modules #

# First party modules #

# Internal modules #
from eu_cbm_hat.combos.reference         import Reference
from eu_cbm_hat.combos.no_market_forcing import NoMarketForcing
from eu_cbm_hat.combos.skewfcth          import Skewfcth
from eu_cbm_hat.combos.reference_crf              import Reference_crf
# from eu_cbm_hat.combos.half_harvest               import Half_harvest
# from eu_cbm_hat.combos.plus_20_harvest            import Plus_20_harvest       
from eu_cbm_hat.combos.potencia_baseline import Potencia_baseline
from eu_cbm_hat.combos.potencia_necp_bme_up100 import Potencia_necp_bme_up100
from eu_cbm_hat.combos.potencia_necp_bme_up200 import Potencia_necp_bme_up200
from eu_cbm_hat.combos.potencia_necp_bms_down50 import Potencia_necp_bms_down50
from eu_cbm_hat.combos.potencia_necp_bms_down90 import Potencia_necp_bms_down90
from eu_cbm_hat.combos.ia_2040 import IA_2040
from eu_cbm_hat.combos.pikbau import PIKBau
from eu_cbm_hat.combos.pikfair import PIKFair

###############################################################################
# List all combo classes to be loaded #
combo_classes = [Reference,
                 NoMarketForcing,
                 Reference_crf,
                 Skewfcth,
                 Potencia_baseline,
#                  Half_harvest,
#                  Plus_20_harvest,
                 Potencia_necp_bme_up100,
                 Potencia_necp_bme_up200,
                 Potencia_necp_bms_down50,
                 Potencia_necp_bms_down90,
                 IA_2040,
                 PIKBau,
                 PIKFair,
                 ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to run the imaginary `ZZ` country to test the pipeline.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/repos/eu_cbm/eu_cbm_hat/scripts/running/run_zz.py
"""

# Internal modules #
from eu_cbm_hat.core.continent import continent
runner = continent.combos['reference'].runners['ZZ'][-1]
# runner.country.base_year = 2020
runner.num_timesteps = 30
runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

# Load output files
# Input events sent to libcbm before hat
events_input = runner.input_data["events"]
# Output events generated by hat and sent to libcbm
events_hat = runner.output.load("events", with_clfrs=False)
# Annual overview of availability
extras = runner.output.load("extras", with_clfrs=False)
# Pools and fluxes
pool_flux = runner.output.pool_flux

# events_input.dist_type_name.unique()
# array([ 8,  7,  5, 12, 14, 16, 22, 11, 21, 42,  4])

# events_hat.dist_type_name.unique()
# array([14, 16, 22, 21, 12, 18, 11, 29])

# pool_flux.disturbance_type.unique()
# array([ 0,  7,  9, 13, 10, 12,  4,  8,  6, 11,  5, 16,  3], dtype=int32)

# Expected provided
runner.qaqc.expected_provided.by(index=["forest_type", "disturbance_type"])






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
from tqdm import tqdm

# First party modules #

# Internal modules #
from eu_cbm_hat.core.continent import continent
from activities_creation import MakeActivities

###############################################################################
makers = [MakeActivities(c) for c in continent]

if __name__ == '__main__':
    print([maker.save_interface() for maker in tqdm(makers)])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

A script to run Luxembourg.

Typically you would run this file from a command line like this:

     ipython3 -i -- ~/deploy/eu_cbm_hat/scripts/running/run_lu.py

"""

from eu_cbm_hat.core.continent import continent


#############################################
# Declare which scenario combination to run #
#############################################
# Scenario combination defined in the yaml file 
# `~/repos/eu_cbm_data/combos/harvest_test.yaml`
# Scenario itself is defined as code in 
# `~/repos/eu_cbm_hat/eu_cbm_hat/combos/harvest_test.py`
combo   = continent.combos['hat']
runner  = combo.runners['LU'][-1]
runner.num_timesteps = 25

# Run the model
output = runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

# Input events sent to libcbm
events_input = runner.input_data["events"]
# Events stored in the output
events_output = runner.output.events
output_extras = runner.output.extras





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""
import warnings
import pandas
from pandas.errors import EmptyDataError

class ExpectedProvided:
    """
    Check harvest expected versus provided

    Usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['hat'].runners["ZZ"][-1]
        >>> # All events including input events and HAT events
        >>> runner.qaqc.expected_provided.events

    Comparison between harvest expected and provided with
    forest type and disturbance type as grouping variables:

        >>> runner.qaqc.expected_provided.by(index = ["forest_type", "disturbance_type"])

    Example with all classifiers as grouping variables:

        >>> clfrs = list(runner.country.orig_data.classif_names.values())
        >>> runner.qaqc.expected_provided.by(index = clfrs + ["disturbance_type"])
    """

    def __init__(self, qaqc):
        # Default attributes #
        self.runner = qaqc.runner

    @property
    def events(self):
        """All events including input events mostly natural disturbances and hat events
        """
        # Load input disturbances, available here after a model run
        events_input = self.runner.input_data["events"]
        # Add year
        events_input["year"] = self.runner.country.timestep_to_year(events_input["step"])

        # Note previous approach checked the path to check if there are HAT
        # related events but it doesn't work because if there are no HAT events
        # an empty file is written at that location
        # path_events_hat = self.runner.output.paths["events"]
        # if not path_events_hat.exists:
        try:
            # Load output disturbances related to HAT
            events_hat = self.runner.output["events"]
            events_hat["step"] = self.runner.country.year_to_timestep(events_hat["year"])
        except ValueError:
            path_events_hat = self.runner.output.paths["events"]
            warnings.warn(f"There are no events related to HAT in: \n{path_events_hat}")
            # In the absence of HAT events, return only the events_input
            return events_input

        # Concatenate the events files
        events_cols = ['status', 'forest_type', 'region', 'mgmt_type', 'mgmt_strategy',
                       'climate', 'con_broad', 'site_index', 'growth_period',
                       'using_id', 'sw_start', 'sw_end', 'hw_start', 'hw_end',
                       'min_since_last_dist', 'max_since_last_dist', 'last_dist_id',
                       'amount', 'dist_type_name', 'measurement_type', 'efficiency',
                       'sort_type', 'step', 'year']
        events = pandas.concat([events_input[events_cols], events_hat[events_cols]])
        return events

    def by(self, index):
        """Expected provided by the given classifiers and disturbance id

        :param list classif: list of classifiers to be used as grouping variables

        The diff only makes sense for disturbance of type M, that is why it is call diff_m.
        For disturbances expressed in terms of area, the amount requested is expressed in area
        and the amount provided in terms of tons of carbon to the products pool.

        Example use with only forest type and disturbance type:

            >>> runner.qaqc.expected_provided.by(index = ["forest_type", "disturbance_type"])

        Use all classifiers as grouping variables:

            >>> clfrs = list(runner.country.orig_data.classif_names.values())
            >>> runner.qaqc.expected_provided.by(index = clfrs + ["disturbance_type"])

        """
        # Load events and pool_flux data
        events = self.events.copy()
        events["disturbance_type"] = events["dist_type_name"]
        pool_flux = self.runner.output.pool_flux

        # Add "year" to the index
        index = index + ["year"]
        # Check that the index is present in both datasets
        missing_events_cols = set(index) - set(events.columns)
        missing_pf_cols = set(index) - set(pool_flux.columns)
        if missing_events_cols:
            msg = "The following columns are not present in the "
            raise ValueError(msg + f"events data \n{missing_events_cols}")
        if missing_pf_cols:
            raise ValueError(msg + f"pool_flux data \n{missing_pf_cols}")

        # Make sure "site_index" is of type character if present
        if "site_index" in index:
            events["site_index"] = events["site_index"].astype(str)

        # Aggregate the events table
        events["measurement_type"] = "amount_" + events["measurement_type"].str.lower()
        events_agg = (events
                      .groupby(index + ["measurement_type"])
                      .agg(amount = ("amount", sum))
                      .reset_index()
                      # Reshape measurement type in columns
                      .pivot(index = index, columns="measurement_type", values="amount")
                     )
        # Add start and end age information
        events_agg["sw_start_min"] = events.groupby(index)["sw_start"].agg(min)
        events_agg["sw_end_max"] = events.groupby(index)["sw_end"].agg(max)

        # Make index columns available for merging
        events_agg = events_agg.reset_index()

        # Aggregate the pool_flux table
        product_cols = ['softwood_merch_to_product', 'softwood_other_to_product',
                        'softwood_stem_snag_to_product', 'softwood_branch_snag_to_product',
                        'hardwood_merch_to_product', 'hardwood_other_to_product',
                        'hardwood_stem_snag_to_product', 'hardwood_branch_snag_to_product']
        flux_agg = pool_flux.groupby(index)[product_cols].agg(sum)
        flux_agg["sum_flux_to_product"] = flux_agg.sum(axis=1)
        # Add age information
        flux_agg["age_min"] = pool_flux.groupby(index)["age"].agg(min)
        flux_agg["age_max"] = pool_flux.groupby(index)["age"].agg(max)
        # Add the area
        flux_agg["area"] = pool_flux.groupby(index)["area"].agg(sum)
        flux_agg = flux_agg.reset_index()

        # Assert that there are no repetition in the index columns
        # It could be the case if additional variables have been added to the index
        assert len(events_agg.set_index(index).index.unique()) == len(events_agg)
        assert len(flux_agg.set_index(index).index.unique()) == len(flux_agg)

        # Merge tables on the index
        # Full outer merge so we don't loose lines
        df = events_agg.merge(flux_agg, how="outer", on=index)

        # Compute the diff
        df["diff_m"] = df["amount_m"] - df["sum_flux_to_product"]
        df["diff_a"] = df["amount_a"] - df["area"]

        # Reorder columns
        cols =  events_agg.columns.to_list()
        cols += ['area', 'diff_a', 'sum_flux_to_product', 'diff_m',
                 'age_min', 'age_max',
                 'softwood_merch_to_product',
                 'softwood_other_to_product', 'softwood_stem_snag_to_product',
                 'softwood_branch_snag_to_product', 'hardwood_merch_to_product',
                 'hardwood_other_to_product', 'hardwood_stem_snag_to_product',
                 'hardwood_branch_snag_to_product']
        assert set(df.columns).issubset(cols)

        return df[cols]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# First party modules #
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.combos.base_combo import Combination
from eu_cbm_hat.cbm.dynamic       import DynamicRunner

###############################################################################
class HarvestTest(Combination):
    """
    An integration test for the dynamic creation of disturbances during
    the model run. Especially aimed at country `ZZ`.
    """

    short_name = 'harvest_test'

    @property_cached
    def runners(self):
        """
        A dictionary of country codes as keys with a list of runners as
        values.
        """
        return {c.iso2_code: [DynamicRunner(self, c, 0)]
                for c in self.continent}
